# -*- coding: utf-8 -*-
# Copyright 2026, CS GROUP - France, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from collections import UserList
from typing import Any
from unittest import mock

import shapely
from shapely import Polygon, geometry
from shapely.errors import ShapelyError

from eodag.api.product import EOProduct
from eodag.api.search_result import SearchResult
from eodag.crunch import (
    FilterDate,
    FilterLatestByName,
    FilterLatestIntersect,
    FilterOverlap,
    FilterProperty,
)
from eodag.plugins.crunch.base import Crunch
from eodag.utils import DEFAULT_SHAPELY_GEOMETRY
from eodag.utils.exceptions import ValidationError


class TestPluginCrunch(unittest.TestCase):

    __product_id: int = 0

    def __fake_search_result(
        self, products_properties: list[dict[str, Any]]
    ) -> SearchResult:
        """Mock search result"""
        search_results = []
        for product_properties in products_properties:
            TestPluginCrunch.__product_id += 1
            properties = {
                "id": "fake_{}".format(TestPluginCrunch.__product_id),
                "geometry": DEFAULT_SHAPELY_GEOMETRY,
            }
            for key in product_properties:
                properties[key] = product_properties[key]

            search_results.append(
                EOProduct(
                    provider="fake_provider",
                    properties=properties,
                    collection="fake_collection",
                )
            )

        return SearchResult(UserList(search_results))

    def test_crunch_base(self):
        """Crunch base test"""
        crunch = Crunch({})
        products: list[EOProduct] = []
        try:
            crunch.proceed(products)
            self.fail("Abstract class Crunch must not be instantiable")
        except NotImplementedError:
            pass
        except Exception as error:
            self.fail(
                "Unexpected error when try to instanciate abstract class Crunch: {error}".format(
                    error=error
                )
            )

    def test_crunch_filterdate(self):
        """Crunch FilterDate test"""
        # No products to filter
        search_results: SearchResult = self.__fake_search_result([])
        filtered_result = search_results.crunch(
            FilterDate(dict(start="2025-01-15", end="2025-01-15"))
        )
        self.assertEqual(len(filtered_result), 0)

        # Some products to filter
        search_results = self.__fake_search_result(
            [
                {"start_datetime": "2025-01-16", "end_datetime": "2025-01-16"},
                {"start_datetime": "2025-01-16", "end_datetime": "2025-01-17"},
                {"start_datetime": "2025-01-17", "end_datetime": "2025-01-17"},
                {"start_datetime": "2025-01-17", "end_datetime": "2025-01-18"},
                {"start_datetime": "2025-01-17", "end_datetime": "2025-01-19"},
            ]
        )
        self.assertEqual(len(search_results), 5)

        # Full filter

        filtered_result = search_results.crunch(
            FilterDate(dict(start="2025-01-15", end="2025-01-15"))
        )
        self.assertEqual(len(filtered_result), 0)

        filtered_result = search_results.crunch(
            FilterDate(dict(start="2025-01-16 00:00:00", end="2025-01-16 23:59:59"))
        )
        self.assertEqual(len(filtered_result), 1)

        filtered_result = search_results.crunch(
            FilterDate(dict(start="2025-01-16", end="2025-01-16"))
        )
        self.assertEqual(len(filtered_result), 1)

        filtered_result = search_results.crunch(
            FilterDate(dict(start="2025-01-16", end="2025-01-17"))
        )

        self.assertEqual(len(filtered_result), 3)

        filtered_result = search_results.crunch(
            FilterDate(dict(start="2025-01-16", end="2025-01-18"))
        )
        self.assertEqual(len(filtered_result), 4)

        # Partial filter

        filtered_result = search_results.crunch(FilterDate(dict(start="2025-01-17")))
        self.assertEqual(len(filtered_result), 3)

        filtered_result = search_results.crunch(FilterDate(dict(end="2025-01-18")))
        self.assertEqual(len(filtered_result), 4)

        # No filters
        filtered_result = search_results.crunch(FilterDate())
        self.assertEqual(len(filtered_result), 5)

        # Wrong filter (invalid date order)
        filtered_result = search_results.crunch(
            FilterDate(dict(start="2025-01-17", end="2025-01-16"))
        )
        self.assertEqual(len(filtered_result), 0)

        # Invalid filter
        try:
            filtered_result = search_results.crunch(
                FilterDate(dict(start="invalid_date", end="2025-01-18"))
            )
            self.fail("Must not let pass invalid date format")
        except ValidationError:
            pass

        try:
            filtered_result = search_results.crunch(
                FilterDate(dict(start="2025-01-16", end="wrong_date"))
            )
            self.fail("Must not let pass invalid date format")
        except ValidationError:
            pass

        # Uncomplete products
        search_results = self.__fake_search_result(
            [
                {"start_datetime": None, "end_datetime": "2025-01-16"},
                {"start_datetime": "2025-01-16", "end_datetime": "2025-01-17"},
                {"start_datetime": "2025-01-17", "end_datetime": None},
                {"start_datetime": "2025-01-17", "end_datetime": "2025-01-18"},
                {"start_datetime": None, "end_datetime": None},
            ]
        )

        filtered_result = search_results.crunch(
            FilterDate(dict(start="2025-01-15", end="2025-01-15"))
        )
        self.assertEqual(len(filtered_result), 1)

        filtered_result = search_results.crunch(FilterDate(dict(start="2025-01-17")))
        self.assertEqual(len(filtered_result), 4)

        filtered_result = search_results.crunch(FilterDate(dict(end="2025-01-18")))
        self.assertEqual(len(filtered_result), 5)

    def test_crunch_filterdate_sort(self):
        """Crunch FilterDate.sort_product_by_start_date test"""

        # sort_product_by_start_date
        products = self.__fake_search_result(
            [
                {"start_datetime": "2025-01-18", "end_datetime": "2025-01-18"},
                {"start_datetime": "2025-01-16", "end_datetime": "2025-01-16"},
                {"start_datetime": "2025-01-17", "end_datetime": "2025-01-17"},
            ]
        )
        products.sort(key=FilterDate.sort_product_by_start_date)
        self.assertEqual(products[0].properties["start_datetime"], "2025-01-16")
        self.assertEqual(products[1].properties["start_datetime"], "2025-01-17")
        self.assertEqual(products[2].properties["start_datetime"], "2025-01-18")

        products.sort(key=FilterDate.sort_product_by_start_date, reverse=True)
        self.assertEqual(products[0].properties["start_datetime"], "2025-01-18")
        self.assertEqual(products[1].properties["start_datetime"], "2025-01-17")
        self.assertEqual(products[2].properties["start_datetime"], "2025-01-16")

        products = self.__fake_search_result(
            [
                {"start_datetime": "2025-01-18", "end_datetime": "2025-01-18"},
                {"start_datetime": None, "end_datetime": "2025-01-16"},
                {"start_datetime": "2025-01-17", "end_datetime": "2025-01-17"},
            ]
        )
        products.sort(key=FilterDate.sort_product_by_start_date, reverse=False)
        self.assertEqual(products[0].properties.get("start_datetime", None), None)
        self.assertEqual(
            products[1].properties.get("start_datetime", None), "2025-01-17"
        )
        self.assertEqual(
            products[2].properties.get("start_datetime", None), "2025-01-18"
        )

    def test_crunch_latest_intersect(self):
        """Crunch FilterLatestIntersect test"""

        # No products to filter
        search_results: SearchResult = self.__fake_search_result([])
        filtered_result = search_results.crunch(FilterLatestIntersect())
        self.assertEqual(len(filtered_result), 0)

        # Some product filter
        search_results = self.__fake_search_result(
            [
                {"geometry": geometry.box(10, 20, 15, 30)},
                {"geometry": geometry.box(10, 30, 15, 40)},
                {"geometry": geometry.box(15, 20, 20, 30)},
                {"geometry": geometry.box(15, 30, 20, 40)},
            ]
        )

        # No filter
        filtered_result = search_results.crunch(FilterLatestIntersect())
        self.assertEqual(len(filtered_result), 4)

        # invalid filter geometry, ignore filter
        filtered_result = search_results.crunch(
            FilterLatestIntersect(), geometry="hell_wrong_geometry"
        )
        self.assertEqual(len(filtered_result), 4)

        # Mismatch filter

        filtered_result = search_results.crunch(
            FilterLatestIntersect(), geometry=geometry.box(0, 0, 0, 0)
        )
        self.assertEqual(len(filtered_result), 0)

        filtered_result = search_results.crunch(
            FilterLatestIntersect(),
            geometry={"lonmin": 0, "latmin": 0, "lonmax": 0, "latmax": 0},
        )
        self.assertEqual(len(filtered_result), 0)

        filtered_result = search_results.crunch(
            FilterLatestIntersect(),
            geometry=Polygon(((0, 0), (0, 1), (1, 1), (1, 0), (0, 0))),
        )
        self.assertEqual(len(filtered_result), 0)

    def test_crunch_latest_intersect_sort(self):
        """Crunch FilterLatestIntersect.sort_product_by_start_date test"""

        # sort_product_by_start_date
        products = self.__fake_search_result(
            [
                {"start_datetime": "2025-01-18", "end_datetime": "2025-01-18"},
                {"start_datetime": "2025-01-16", "end_datetime": "2025-01-16"},
                {"start_datetime": "2025-01-17", "end_datetime": "2025-01-17"},
            ]
        )
        products.sort(key=FilterLatestIntersect.sort_product_by_start_date)
        self.assertEqual(products[0].properties["start_datetime"], "2025-01-16")
        self.assertEqual(products[1].properties["start_datetime"], "2025-01-17")
        self.assertEqual(products[2].properties["start_datetime"], "2025-01-18")

        products.sort(
            key=FilterLatestIntersect.sort_product_by_start_date, reverse=True
        )
        self.assertEqual(products[0].properties["start_datetime"], "2025-01-18")
        self.assertEqual(products[1].properties["start_datetime"], "2025-01-17")
        self.assertEqual(products[2].properties["start_datetime"], "2025-01-16")

        products = self.__fake_search_result(
            [
                {"start_datetime": "2025-01-18", "end_datetime": "2025-01-18"},
                {"start_datetime": None, "end_datetime": "2025-01-16"},
                {"start_datetime": "2025-01-17", "end_datetime": "2025-01-17"},
            ]
        )
        products.sort(
            key=FilterLatestIntersect.sort_product_by_start_date, reverse=False
        )
        self.assertEqual(products[0].properties.get("start_datetime", None), None)
        self.assertEqual(
            products[1].properties.get("start_datetime", None), "2025-01-17"
        )
        self.assertEqual(
            products[2].properties.get("start_datetime", None), "2025-01-18"
        )

    def test_crunch_lastestbyname(self):
        """Crunch FilterLatestByName test"""

        # Missing parameter
        search_results: SearchResult = self.__fake_search_result([])
        try:
            _ = search_results.crunch(FilterLatestByName())
            self.fail('FilterLatestByName require parameter "name_pattern"')
        except ValidationError:
            pass

        # Invalid pattern
        try:
            _ = search_results.crunch(
                FilterLatestByName({"name_pattern": "hell_pattern!"})
            )
            self.fail('FilterLatestByName require valid parameter "name_pattern"')
        except ValidationError:
            pass

        # No products
        filter_products = search_results.crunch(
            FilterLatestByName({"name_pattern": "(?P<tileid>\\d{6})"})
        )
        self.assertEqual(len(filter_products), 0)

        # Some products to filter
        search_results = self.__fake_search_result(
            [
                {"start_datetime": "2025-01-18", "title": "000001"},
                {"start_datetime": "2025-01-18", "title": "000002"},
                {"start_datetime": "2025-01-19", "title": "000003"},
                {"start_datetime": "2025-01-19", "title": "000004"},
                {"start_datetime": "2025-01-20", "title": "000005"},
            ]
        )
        filter_products = search_results.crunch(
            FilterLatestByName({"name_pattern": "(?P<tileid>\\d{6})"})
        )
        self.assertEqual(len(filter_products), 5)

        # Some products to filter with mixed title
        search_results = self.__fake_search_result(
            [
                {"start_datetime": "2025-01-18", "title": "malformed"},
                {"start_datetime": "2025-01-18", "title": "000002"},
                {"start_datetime": "2025-01-19", "title": "malformed"},
                {"start_datetime": "2025-01-19", "title": "000004"},
                {"start_datetime": "2025-01-20", "title": "malformed"},
            ]
        )
        filter_products = search_results.crunch(
            FilterLatestByName({"name_pattern": "(?P<tileid>\\d{6})"})
        )
        self.assertEqual(len(filter_products), 2)

        # Some products to filter with mixed title, repeated formatted title
        search_results = self.__fake_search_result(
            [
                {"start_datetime": "2025-01-18", "title": "malformed"},
                {"start_datetime": "2025-01-18", "title": "000002"},
                {"start_datetime": "2025-01-19", "title": "malformed"},
                {"start_datetime": "2025-01-19", "title": "000002"},
                {"start_datetime": "2025-01-20", "title": "malformed"},
            ]
        )
        filter_products = search_results.crunch(
            FilterLatestByName({"name_pattern": "(?P<tileid>\\d{6})"})
        )
        self.assertEqual(len(filter_products), 1)

    def test_crunch_overlap(self):
        """Crunch FilterOverlap test"""

        # No products, no configuration
        search_results: SearchResult = self.__fake_search_result([])
        filtered_result = search_results.crunch(FilterOverlap())
        self.assertEqual(len(filtered_result), 0)

        filtered_result = search_results.crunch(
            FilterOverlap(), geometry=geometry.box(0, 0, 0, 0)
        )
        self.assertEqual(len(filtered_result), 0)

        # Invalid configuration
        search_results = self.__fake_search_result(
            [
                {"geometry": geometry.box(10, 20, 15, 30)},
                {"geometry": geometry.box(10, 30, 15, 40)},
                {"geometry": geometry.box(15, 20, 20, 30)},
                {"geometry": geometry.box(15, 30, 20, 40)},
            ]
        )
        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": -1,  # min: 0
                    "contains": False,
                    "intersects": False,
                    "within": False,
                }
            ),
            geometry=geometry.box(0, 0, 0, 0),
        )
        self.assertEqual(len(filtered_result), 0)

        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": "hell_minimum!",
                    "contains": False,
                    "intersects": False,
                    "within": False,
                }
            ),
            geometry=geometry.box(0, 0, 0, 0),
        )
        self.assertEqual(len(filtered_result), 0)

        # Parameters exclusion (contains, intersects, within)
        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": 100,
                    "contains": True,
                    "intersects": True,
                    "within": False,
                }
            ),
            geometry=geometry.box(0, 0, 0, 0),
        )
        self.assertEqual(len(filtered_result), 4)

        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": 100,
                    "contains": True,
                    "intersects": False,
                    "within": True,
                }
            ),
            geometry=geometry.box(0, 0, 0, 0),
        )
        self.assertEqual(len(filtered_result), 4)

        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": 100,
                    "contains": False,
                    "intersects": True,
                    "within": True,
                }
            ),
            geometry=geometry.box(0, 0, 0, 0),
        )
        self.assertEqual(len(filtered_result), 4)

        # Intersect
        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": 95,
                    "contains": False,
                    "intersects": True,
                    "within": False,
                }
            ),
            geometry=geometry.box(12, 25, 13, 35),
        )
        self.assertEqual(len(filtered_result), 2)

        # Contains
        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": 100,
                    "contains": True,
                    "intersects": False,
                    "within": False,
                }
            ),
            geometry=geometry.box(12, 25, 13, 35),
        )
        self.assertEqual(len(filtered_result), 0)

        # Contains
        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": 100,
                    "contains": True,
                    "intersects": False,
                    "within": False,
                }
            ),
            geometry=geometry.box(12, 21, 13, 29),
        )
        self.assertEqual(len(filtered_result), 1)

        # Within
        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": 100,
                    "contains": False,
                    "intersects": False,
                    "within": True,
                }
            ),
            geometry=geometry.box(9, 19, 16, 31),
        )
        self.assertEqual(len(filtered_result), 1)

        # Search geom area = 0
        geom = Polygon(((10, 20), (15, 30), (15, 20), (10, 30), (10, 20)))
        self.assertFalse(geom.is_valid)
        search_results = self.__fake_search_result(
            [
                {"geometry": geometry.box(10, 20, 15, 30)},
                {"geometry": geometry.box(10, 30, 15, 40)},
                {"geometry": geometry.box(15, 20, 20, 30)},
                {"geometry": geometry.box(15, 30, 20, 40)},
            ]
        )
        filtered_result = search_results.crunch(
            FilterOverlap(
                {
                    "minimum_overlap": 50,
                    "contains": False,
                    "intersects": True,
                    "within": False,
                }
            ),
            geometry=geometry.box(10, 30, 10, 30),
        )
        self.assertEqual(len(filtered_result), 0)

        # Product invalid geometry (butterfly shape) recoverable
        geom = Polygon(((10, 20), (15, 30), (15, 20), (10, 30), (10, 20)))
        self.assertFalse(geom.is_valid)
        search_results = self.__fake_search_result([{"geometry": geom}])
        filter = FilterOverlap(
            {
                "minimum_overlap": 50,
                "contains": False,
                "intersects": True,
                "within": False,
            }
        )

        # Disable search intersection to test invalid geom
        for index in range(0, len(search_results)):
            search_results[index].search_intersection = None

        filtered_result = search_results.crunch(
            filter, geometry=geometry.box(10, 30, 20, 40)
        )
        self.assertEqual(len(filtered_result), 1)

        # Force fail search_geom intersection
        def shapely_intersect_forced_exception(a, b, grid_size=None, **kwargs):
            print("Mocked: no shapely.intersection allowed")
            raise ShapelyError("No intersection allowed")

        with mock.patch.object(
            shapely, "intersection", new=shapely_intersect_forced_exception
        ):
            filtered_result = search_results.crunch(
                filter, geometry=geometry.box(10, 30, 20, 40)
            )
            self.assertEqual(len(filtered_result), 0)

    def test_crunch_property(self):
        """Crunch FilterProperty test"""

        # No configuration
        search_results = self.__fake_search_result(
            [
                {"myproperty": 1},
                {"myproperty": 2},
                {"myproperty": 3},
                {"myproperty": 4},
                {"myproperty": 5},
            ]
        )
        filtered_result = search_results.crunch(FilterProperty({}))
        self.assertEqual(len(filtered_result), 5)

        # Partial configuration
        filtered_result = search_results.crunch(FilterProperty({"myproperty": 3}))
        self.assertEqual(len(filtered_result), 1)

        filtered_result = search_results.crunch(FilterProperty({"myproperty": 6}))
        self.assertEqual(len(filtered_result), 0)

        # Wrong configuration
        filtered_result = search_results.crunch(
            FilterProperty({"myproperty": 6, "operator": "hell"})
        )
        self.assertEqual(len(filtered_result), 5)

        # Full configuration
        for test in [
            {"value": 2, "operator": "lt", "expect_results": 1},
            {"value": 2, "operator": "le", "expect_results": 2},
            {"value": 2, "operator": "eq", "expect_results": 1},
            {"value": 2, "operator": "ne", "expect_results": 4},
            {"value": 2, "operator": "ge", "expect_results": 4},
            {"value": 2, "operator": "gt", "expect_results": 3},
            {"value": 4, "operator": "lt", "expect_results": 3},
            {"value": 4, "operator": "le", "expect_results": 4},
            {"value": 4, "operator": "eq", "expect_results": 1},
            {"value": 4, "operator": "ne", "expect_results": 4},
            {"value": 4, "operator": "ge", "expect_results": 2},
            {"value": 4, "operator": "gt", "expect_results": 1},
        ]:
            filtered_result = search_results.crunch(
                FilterProperty(
                    {"myproperty": test["value"], "operator": test["operator"]}
                )
            )
            self.assertEqual(len(filtered_result), test["expect_results"])

        # Multitypes data
        search_results = self.__fake_search_result(
            [
                {"myproperty": 1},
                {"myproperty": "2"},
                {"myproperty": bool},
                {"myproperty": None},
                {"myproperty": 5},
            ]
        )
        filtered_result = search_results.crunch(FilterProperty({"myproperty": 3}))
        self.assertEqual(len(filtered_result), 0)

        filtered_result = search_results.crunch(FilterProperty({"myproperty": 2}))
        self.assertEqual(len(filtered_result), 0)

        filtered_result = search_results.crunch(FilterProperty({"myproperty": "2"}))
        self.assertEqual(len(filtered_result), 1)

        filtered_result = search_results.crunch(FilterProperty({"myproperty": None}))
        self.assertEqual(len(filtered_result), 1)

        # Mismatch filter value type makes no filter wihtout error
        filtered_result = search_results.crunch(FilterProperty({"myproperty": True}))
        self.assertEqual(len(filtered_result), 1)

        # Multitypes data with operator set
        for value in [42, "myvalue", True]:
            for test in [
                {"operator": "lt", "expect_match": False},  # Lower (False)
                {"operator": "le", "expect_match": True},  # Lower(False) + Equal(True)
                {"operator": "eq", "expect_match": True},  # Equal(True)
                {"operator": "ne", "expect_match": False},  # Non Equal(False)
                {
                    "operator": "ge",
                    "expect_match": True,
                },  # Greater(False) + Equal(True)
                {"operator": "gt", "expect_match": False},  # Greater(False)
            ]:
                search_results = self.__fake_search_result([{"myproperty": value}])
                filtered_result = search_results.crunch(
                    FilterProperty({"myproperty": value, "operator": test["operator"]})
                )
                self.assertEqual(len(filtered_result) == 1, test["expect_match"])



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EODAG drivers package"""
from __future__ import annotations

from typing import Callable

from typing_extensions import TypedDict

from .generic import DatasetDriver, GenericDriver
from .sentinel1 import Sentinel1Driver
from .sentinel2 import Sentinel2Driver


class DriverCriteria(TypedDict):
    """Driver criteria definition"""

    #: Function that returns True if the driver is suitable for the given :class:`~eodag.api.product._product.EOProduct`
    criteria: list[Callable[..., bool]]
    #: driver to use
    driver: DatasetDriver


#: list of drivers and their criteria
DRIVERS: list[DriverCriteria] = [
    {
        "criteria": [lambda product: Sentinel2Driver.match(product, by="collection")],
        "driver": Sentinel2Driver(),
    },
    {
        "criteria": [lambda product: Sentinel1Driver.match(product, by="collection")],
        "driver": Sentinel1Driver(),
    },
    {
        "criteria": [lambda product: Sentinel2Driver.match(product, by="properties")],
        "driver": Sentinel2Driver(),
    },
    {
        "criteria": [lambda product: Sentinel1Driver.match(product, by="properties")],
        "driver": Sentinel1Driver(),
    },
    {
        "criteria": [lambda product: GenericDriver.match(product)],
        "driver": GenericDriver(),
    },
]

# exportable content
__all__ = ["DRIVERS", "DatasetDriver", "GenericDriver", "Sentinel2Driver"]



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EODAG Command Line interface

Usage: eodag [OPTIONS] COMMAND [ARGS]...

  Earth Observation Data Access Gateway: work on EO products from any
  provider

Options:
  -v, --verbose  Control the verbosity of the logs. For maximum verbosity,
                 type -vvv
  --help         Show this message and exit.

Commands:
  discover         Fetch providers to discover collections
  download         Download a list of products from a serialized search...
  list             List supported collections
  search           Search satellite images by their collections,...
  version          Print eodag version and exit

  noqa: D103
"""

from __future__ import annotations

import functools
import json
import sys
import textwrap
from importlib.metadata import metadata
from typing import TYPE_CHECKING, Any, Callable, Mapping, Optional
from urllib.parse import parse_qs

import click

from eodag.utils import DEFAULT_LIMIT, DEFAULT_PAGE

if TYPE_CHECKING:
    from click import Context

# A list of supported crunchers that the user can choose (see --cruncher option below)
CRUNCHERS = [
    "FilterLatestByName",
    "FilterLatestIntersect",
    "FilterOverlap",
    "FilterProperty",
    "FilterDate",
]


class MutuallyExclusiveOption(click.Option):
    """Mutually Exclusive Options for Click
    from https://gist.github.com/jacobtolar/fb80d5552a9a9dfc32b12a829fa21c0c
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs["help"] = help + (
                " NOTE: This argument is mutually exclusive with "
                " arguments: [" + ex_str + "]."
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(
        self, ctx: Context, opts: Mapping[str, Any], args: list[str]
    ):
        """Raise error or use parent handle_parse_result()"""
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise click.UsageError(
                "Illegal usage: `{}` is mutually exclusive with arguments `{}`.".format(
                    self.name, ", ".join(self.mutually_exclusive)
                )
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(ctx, opts, args)


def _deprecated_cli(message: str, version: Optional[str] = None) -> Callable[..., Any]:
    """Decorator to mark a CLI command as deprecated and print a bold yellow warning."""
    version_msg = f" -- Deprecated since v{version}" if version else ""

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            full_message = f"DEPRECATED: {message}{version_msg}"
            click.echo(click.style(full_message, fg="yellow", bold=True), err=True)
            return func(*args, **kwargs)

        return wrapper

    return decorator


@click.group(chain=True)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Control the verbosity of the logs. For maximum verbosity, type -vvv",
)
@click.pass_context
def eodag_cli(ctx: Context, verbose: int) -> None:
    """Earth Observation Data Access Gateway: work on EO products from any provider"""
    if ctx.obj is None:
        ctx.obj = {}
    ctx.obj["verbosity"] = verbose


@eodag_cli.command(name="version", help="Print eodag version and exit")
def version() -> None:
    """Print eodag version and exit"""
    click.echo(
        "{__title__} ({__description__}): version {__version__}".format(
            __title__=metadata("eodag")["Name"],
            __description__=metadata("eodag")["Summary"],
            __version__=metadata("eodag")["Version"],
        )
    )


@eodag_cli.command(
    name="search",
    help="Search satellite images by their collections, instruments, constellation, "
    "platform, processing level or sensor type. It is mandatory to provide "
    "at least one of the previous criteria for eodag to perform a search. "
    "Optionally crunch the search results before storing them in a geojson file",
)
@click.option(
    "-f",
    "--conf",
    help="File path to the user configuration file with its credentials, default is ~/.config/eodag/eodag.yml",
    type=click.Path(exists=True),
)
@click.option(
    "-l",
    "--locs",
    help="File path to the user locations configuration file, default is ~/.config/eodag/locations.yml",
    type=click.Path(exists=True),
)
@click.option("-p", "--provider", help="Search on this provider")
@click.option(
    "-b",
    "--box",
    type=(float,) * 4,
    default=(None,) * 4,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["geom"],
    help="Search for a product on a bounding box, providing its minlon, minlat, "
    "maxlon and maxlat (in this order).",
)
@click.option(
    "-g",
    "--geom",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["box"],
    help="Search for a product on a geometry, providing its WKT representation.",
)
@click.option(
    "-s",
    "--start",
    type=click.DateTime(),
    help="Start sensing time in ISO8601 format (e.g. '1990-11-26', '1990-11-26T14:30:10'). UTC is assumed",
)
@click.option(
    "-e",
    "--end",
    type=click.DateTime(),
    help="End sensing time in ISO8601 format (e.g. '1990-11-26', '1990-11-26T14:30:10'). UTC is assumed",
)
@click.option("-c", "--collection", help="The collection to search")
@click.option("--instruments", help="Search for products matching these instruments")
@click.option("--platform", help="Search for products matching this platform")
@click.option("--constellation", help="Search for products matching this constellation")
@click.option(
    "--processing-level", help="Search for products matching this processing level"
)
@click.option("--sensor-type", help="Search for products matching this type of sensor")
@click.option(
    "--cloud-cover",
    type=click.IntRange(0, 100),
    help="Maximum cloud cover percentage needed for the product",
)
@click.option("--id", help="Search for the product identified by this id")
@click.option(
    "--locations",
    type=str,
    help="Custom query-string argument(s) to select locations. "
    "Format :'key1=value1&key2=value2'. Example: --locations country=FRA&continent=Africa",
)
@click.option(
    "-q",
    "--query",
    type=str,
    help="Custom query-string argument(s). Format :'key1=value1&key2=value2'",
)
@click.option(
    "--cruncher",
    type=click.Choice(CRUNCHERS),
    multiple=True,
    help="A cruncher to be applied to search results. Repeat option many times to "
    "apply many crunchers",
)
@click.option(
    "--cruncher-args",
    type=(str,) * 3,
    multiple=True,
    help="Named arguments acting as the parameters of a cruncher. "
    "Enter it like this: --cruncher-args <CruncherName> <arg-name> <arg-value>. "
    "Repeat option many times to give many args to a cruncher",
)
@click.option(
    "--storage",
    type=click.Path(dir_okay=False, writable=True, readable=False),
    default="search_results.geojson",
    help="Path to the file where to store search results (.geojson extension will be "
    "automatically appended to the filename). [default: search_results.geojson]",
)
@click.option(
    "--items",
    type=int,
    show_default=False,
    help="(DEPRECATED, use limit instead) The number of items to return. Eodag is bound to "
    "whatever limitation the providers have on the number of results they return. This option "
    "allows to control how many items eodag should request "
    f"[default: {DEFAULT_LIMIT}]",
)
@click.option(
    "--limit",
    type=int,
    show_default=False,
    help="The number of items to return. Eodag is bound to whatever limitation the "
    "providers have on the number of results they return. This option allows "
    "to control how many items eodag should request "
    f"[default: {DEFAULT_LIMIT}]",
)
@click.option(
    "--page",
    type=int,
    default=DEFAULT_PAGE,
    show_default=True,
    help="Retrieve the given page",
)
@click.option(
    "--all",
    is_flag=True,
    help="Retrieve ALL the products that match the search criteria. It collects "
    "products by iterating over the results pages until no more products are available."
    "At each iteration, the maximum number of items searched is either 'items' if set, "
    "or a maximum value defined internally for the requested provider, or a default "
    "maximum value equals to 50.",
)
@click.option(
    "--count",
    is_flag=True,
    help="Make a count request together with search (Enabling count will significantly "
    "slow down search requests for some providers, and might be unavailable for some"
    "others).",
)
@click.pass_context
def search_crunch(ctx: Context, **kwargs: Any) -> None:
    """Search collections and optionnaly apply crunchers to search results"""
    # Process inputs for search
    provider = kwargs.pop("provider")
    collection = kwargs.pop("collection")
    instruments = kwargs.pop("instruments")
    platform = kwargs.pop("platform")
    constellation = kwargs.pop("constellation")
    processing_level = kwargs.pop("processing_level")
    sensor_type = kwargs.pop("sensor_type")
    id_ = kwargs.pop("id")
    locations_qs = kwargs.pop("locations")
    custom = kwargs.pop("query")
    if not any(
        [
            collection,
            instruments,
            platform,
            constellation,
            processing_level,
            sensor_type,
            id_,
        ]
    ):
        with click.Context(search_crunch) as ctx:
            print("Give me some work to do. See below for how to do that:", end="\n\n")
            click.echo(search_crunch.get_help(ctx))
        sys.exit(-1)

    from eodag.api.core import EODataAccessGateway
    from eodag.utils.dates import to_iso_utc_string
    from eodag.utils.logging import setup_logging

    setup_logging(verbose=ctx.obj["verbosity"])

    if kwargs["box"] != (None,) * 4:
        rect = kwargs.pop("box")
        footprint = {
            "lonmin": rect[0],
            "latmin": rect[1],
            "lonmax": rect[2],
            "latmax": rect[3],
        }
    else:
        footprint = kwargs.pop("geom")

    start_date = kwargs.pop("start")
    stop_date = kwargs.pop("end")
    criteria = {
        "provider": provider,
        "geometry": footprint,
        "start_datetime": None,
        "end_datetime": None,
        "eo:cloud_cover": kwargs.pop("cloud_cover"),
        "collection": collection,
        "instruments": instruments,
        "constellation": constellation,
        "platform": platform,
        "processing:level": processing_level,
        "eodag:sensor_type": sensor_type,
        "id": id_,
    }
    if custom:
        custom_dict = parse_qs(custom)
        for k, v in custom_dict.items():
            if isinstance(v, list) and len(v) == 1:
                criteria[k] = v[0]
            else:
                criteria[k] = v
    if locations_qs is not None:
        locations = {key: val[0] for key, val in parse_qs(locations_qs).items()}
    else:
        locations = None
    criteria["locations"] = locations
    if start_date:
        criteria["start_datetime"] = to_iso_utc_string(start_date)
    if stop_date:
        criteria["end_datetime"] = to_iso_utc_string(stop_date)
    conf_file = kwargs.pop("conf")
    if conf_file:
        conf_file = click.format_filename(conf_file)
    locs_file = kwargs.pop("locs")
    if locs_file:
        locs_file = click.format_filename(locs_file)

    count = kwargs.pop("count")

    # Process inputs for crunch
    cruncher_names: set[Any] = set(kwargs.pop("cruncher") or [])
    cruncher_args = kwargs.pop("cruncher_args")
    cruncher_args_dict: dict[str, dict[str, Any]] = {}
    if cruncher_args:
        for cruncher, argname, argval in cruncher_args:
            cruncher_args_dict.setdefault(cruncher, {}).setdefault(argname, argval)

    items = kwargs.pop("items")
    if items is not None:
        click.echo(
            click.style(
                "DEPRECATED: --items is deprecated, use --limit instead -- Deprecated since v4.0",
                fg="yellow",
                bold=True,
            ),
            err=True,
        )
    limit = kwargs.pop("limit") or items
    page = kwargs.pop("page") or 1

    gateway = EODataAccessGateway(
        user_conf_file_path=conf_file, locations_conf_path=locs_file
    )

    # Search
    get_all_products = kwargs.pop("all")
    if get_all_products:
        # search_all needs limit to be None if the user lets eodag determines
        # what value it should take.
        limit = None if limit is None else limit
        results = gateway.search_all(limit=limit, **criteria)
    else:
        # search should better take a value that is not None
        limit = DEFAULT_LIMIT if limit is None else limit
        results = gateway.search(count=count, page=page, limit=limit, **criteria)
        if results.number_matched is not None:
            click.echo(
                "Found a total number of {} products".format(results.number_matched)
            )
    click.echo("Returned {} products".format(len(results)))

    # Crunch !
    crunch_args = {
        cruncher_name: cruncher_args_dict.get(cruncher_name, {})
        for cruncher_name in cruncher_names
    }
    if crunch_args:
        results = gateway.crunch(results, search_criteria=criteria, **crunch_args)

    storage_filepath = kwargs.pop("storage")
    if not storage_filepath.endswith(".geojson"):
        storage_filepath += ".geojson"
    result_storage = gateway.serialize(results, filename=storage_filepath)
    click.echo("Results stored at '{}'".format(result_storage))
    ctx.obj["search_results"] = results


@eodag_cli.command(name="list", help="List supported collections")
@click.option("-p", "--provider", help="List collections supported by this provider")
@click.option(
    "--instruments", help="List collections originating from these instruments"
)
@click.option("--platform", help="List collections originating from this platform")
@click.option(
    "--constellation",
    help="List collections originating from this constellation",
)
@click.option("--processing-level", help="List collections of processing level")
@click.option(
    "--sensor-type", help="List collections originating from this type of sensor"
)
@click.option(
    "--no-fetch", is_flag=True, help="Do not fetch providers for new collections"
)
@click.pass_context
def list_col(ctx: Context, **kwargs: Any) -> None:
    """Print the list of supported collections"""
    from eodag.api.collection import CollectionsList
    from eodag.api.core import EODataAccessGateway
    from eodag.utils.exceptions import NoMatchingCollection, UnsupportedProvider
    from eodag.utils.logging import setup_logging

    setup_logging(verbose=ctx.obj["verbosity"])
    dag = EODataAccessGateway()
    provider = kwargs.pop("provider")
    fetch_providers = not kwargs.pop("no_fetch")
    text_wrapper = textwrap.TextWrapper()
    guessed_collections = CollectionsList([])
    try:
        guessed_collections = dag.guess_collection(
            **kwargs,
        )
    except NoMatchingCollection:
        if any(
            kwargs[arg]
            for arg in [
                "instruments",
                "constellation",
                "platform",
                "processing_level",
                "sensor_type",
            ]
        ):
            click.echo("No collection match the following criteria you provided:")
            click.echo(
                "\n".join(
                    "-{param}={value}".format(**locals())
                    for param, value in kwargs.items()
                    if value is not None
                )
            )
            sys.exit(1)
    try:
        if guessed_collections:
            collections = CollectionsList(
                [
                    col
                    for col in dag.list_collections(
                        provider=provider, fetch_providers=fetch_providers
                    )
                    if col.id in [guessed_col.id for guessed_col in guessed_collections]
                ]
            )
        else:
            collections = dag.list_collections(
                provider=provider, fetch_providers=fetch_providers
            )
        click.echo("Listing available collections:")
        for collection in collections:
            click.echo("\n* {}: ".format(collection.id))
            for prop, value in collection.model_dump().items():
                if prop != "id":
                    text_wrapper.initial_indent = "    - {}: ".format(prop)
                    text_wrapper.subsequent_indent = " " * len(
                        text_wrapper.initial_indent
                    )
                    if value is not None:
                        click.echo(text_wrapper.fill(str(value)))
    except UnsupportedProvider:
        click.echo("Unsupported provider. You may have a typo")
        click.echo("Available providers: {}".format(", ".join(dag.providers.names)))
        sys.exit(1)


@eodag_cli.command(name="discover", help="Fetch providers to discover collections")
@click.option("-p", "--provider", help="Fetch only the given provider")
@click.option(
    "--storage",
    type=click.Path(dir_okay=False, writable=True, readable=False),
    default="ext_collections.json",
    help="Path to the file where to store external collections configuration "
    "(.json extension will be automatically appended to the filename). "
    "[default: ext_collections.json]",
)
@click.pass_context
def discover_col(ctx: Context, **kwargs: Any) -> None:
    """Fetch external collections configuration and save result"""
    from eodag.api.core import EODataAccessGateway
    from eodag.utils.logging import setup_logging

    setup_logging(verbose=ctx.obj["verbosity"])
    dag = EODataAccessGateway()
    provider = kwargs.pop("provider")

    ext_collections_conf = (
        dag.discover_collections(provider=provider)
        if provider
        else dag.discover_collections()
    )

    storage_filepath = kwargs.pop("storage")
    if not storage_filepath.endswith(".json"):
        storage_filepath += ".json"
    with open(storage_filepath, "w") as f:
        json.dump(ext_collections_conf, f)
    click.echo("Results stored at '{}'".format(storage_filepath))


@eodag_cli.command(
    help="""Download a list of products from a serialized search result or STAC items URLs/paths

Examples:

  eodag download --search-results /path/to/search_results.geojson

  eodag download --stac-item https://example.com/stac/item1.json --stac-item /path/to/item2.json
""",
)
@click.option(
    "--search-results",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to a serialized search result",
)
@click.option(
    "--stac-item",
    multiple=True,
    help="URL/path of a STAC item to download (multiple values accepted)",
)
@click.option(
    "-f",
    "--conf",
    type=click.Path(exists=True),
    help="File path to the user configuration file with its credentials [default: ~/.config/eodag/eodag.yml]",
)
@click.option(
    "--quicklooks",
    is_flag=True,
    show_default=False,
    help="Download only quicklooks of products instead full set of files",
)
@click.option(
    "--output-dir",
    type=click.Path(dir_okay=True, file_okay=False),
    help="Products or quicklooks download directory [default: local temporary directory]",
)
@click.option(
    "--max-workers",
    type=int,
    help="The maximum number of workers to use for downloading products and assets in parallel",
)
@click.pass_context
def download(ctx: Context, **kwargs: Any) -> None:
    """Download a bunch of products from a serialized search result"""
    search_result_path = kwargs.pop("search_results")
    stac_items = kwargs.pop("stac_item")
    search_results = ctx.obj.get("search_results")
    if not search_result_path and not stac_items and search_results is None:
        with click.Context(download) as ctx:
            click.echo("Nothing to do (no search results file or stac item provided)")
            click.echo(download.get_help(ctx))
        sys.exit(1)

    from concurrent.futures import ThreadPoolExecutor

    from eodag.api.core import EODataAccessGateway
    from eodag.api.search_result import SearchResult
    from eodag.utils.logging import setup_logging

    setup_logging(verbose=ctx.obj["verbosity"])
    conf_file = kwargs.pop("conf")
    if conf_file:
        conf_file = click.format_filename(conf_file)

    satim_api = EODataAccessGateway(user_conf_file_path=conf_file)

    search_results = search_results or SearchResult([])
    if search_result_path:
        search_results.extend(satim_api.deserialize_and_register(search_result_path))
    if stac_items:
        search_results.extend(satim_api.import_stac_items(list(stac_items)))

    output_dir = kwargs.pop("output_dir")
    get_quicklooks = kwargs.pop("quicklooks")

    if get_quicklooks:
        # Download only quicklooks
        click.echo(
            "Flag 'quicklooks' specified, downloading only quicklooks of products"
        )

        for idx, product in enumerate(search_results):
            downloaded_file = product.get_quicklook(output_dir=output_dir)
            if not downloaded_file:
                click.echo(
                    "A quicklook may have been downloaded but we cannot locate it. "
                    "Increase verbosity for more details: `eodag -v download [OPTIONS]`"
                )
            else:
                click.echo("Downloaded {}".format(downloaded_file))

    else:
        # Download products
        executor = ThreadPoolExecutor(max_workers=kwargs.pop("max_workers"))
        downloaded_files = satim_api.download_all(
            search_results, output_dir=output_dir, executor=executor
        )
        if downloaded_files and len(downloaded_files) > 0:
            for downloaded_file in downloaded_files:
                if downloaded_file is None:
                    click.echo(
                        "A file may have been downloaded but we cannot locate it"
                    )
                else:
                    click.echo("Downloaded {}".format(downloaded_file))
        else:
            click.echo(
                "Error during download, a file may have been downloaded but we cannot locate it"
            )


if __name__ == "__main__":
    eodag_cli(obj={})



# -*- coding: utf-8 -*-
# Copyright 2022, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import datetime as dt
import hashlib
import logging
import re
from collections import OrderedDict
from types import MethodType
from typing import TYPE_CHECKING, Annotated, Any, Optional
from urllib.parse import quote_plus, unquote_plus

import geojson
import orjson
from pydantic import AliasChoices
from pydantic.fields import FieldInfo
from requests.auth import AuthBase
from typing_extensions import get_args  # noqa: F401

from eodag.api.product import EOProduct
from eodag.api.product.metadata_mapping import (
    DEFAULT_GEOMETRY,
    NOT_AVAILABLE,
    OFFLINE_STATUS,
    STAGING_STATUS,
    format_metadata,
    mtd_cfg_as_conversion_and_querypath,
    properties_from_json,
)
from eodag.api.search_result import RawSearchResult, SearchResult
from eodag.plugins.search import PreparedSearch
from eodag.plugins.search.qssearch import PostJsonSearch, QueryStringSearch
from eodag.types import json_field_definition_to_python  # noqa: F401
from eodag.types.queryables import Queryables, QueryablesDict
from eodag.types.stac_extensions import EcmwfItemProperties
from eodag.utils import (
    DEFAULT_SEARCH_TIMEOUT,
    deepcopy,
    dict_items_recursive_sort,
    format_string,
    get_geometry_from_ecmwf_area,
    get_geometry_from_ecmwf_feature,
    get_geometry_from_ecmwf_location,
    get_geometry_from_various,
)
from eodag.utils.cache import instance_cached_method
from eodag.utils.dates import (
    COMPACT_DATE_RANGE_PATTERN,
    DATE_RANGE_PATTERN,
    compute_date_range_from_params,
    format_date,
    is_range_in_range,
    parse_date,
    parse_to_utc,
    parse_year_month_day,
    time_values_to_hhmm,
    to_iso_utc_string,
    validate_datetime_param,
)
from eodag.utils.exceptions import (
    DownloadError,
    MisconfiguredError,
    NotAvailableError,
    ValidationError,
)
from eodag.utils.requests import fetch_json

if TYPE_CHECKING:
    from eodag.config import PluginConfig

logger = logging.getLogger("eodag.search.build_search_result")

ECMWF_PREFIX = "ecmwf:"

ALLOWED_KEYWORDS = set(
    [k.replace("ecmwf_", "") for k in EcmwfItemProperties.model_fields.keys()]
)

END = "end_datetime"

START = "start_datetime"


def ecmwf_mtd() -> dict[str, Any]:
    """
    Make metadata mapping dict from a list of defined ECMWF Keywords

    We automatically add the #to_geojson convert to prevent modification of entries by eval() in the metadata mapping.

    keyword:
        - keyword
        - $."keyword"#to_geojson

    :return: metadata mapping dict
    """
    return {k: [k, f'{{$."{k}"#to_geojson}}'] for k in ALLOWED_KEYWORDS}


def _update_properties_from_element(
    prop: dict[str, Any], element: dict[str, Any], values: list[str]
) -> None:
    """updates a property dict with the given values based on the information from the element dict
    e.g. the type is set based on the type of the element
    """
    # multichoice elements are transformed into array
    if element["type"] in ("StringListWidget", "StringListArrayWidget"):
        prop["type"] = "array"
        if values:
            prop["items"] = {"type": "string", "enum": sorted(values)}

    # single choice elements are transformed into string
    elif element["type"] in (
        "StringChoiceWidget",
        "DateRangeWidget",
        "FreeformInputWidget",
    ):
        prop["type"] = "string"
        if values:
            prop["enum"] = sorted(values)

    # a bbox element
    elif element["type"] in ["GeographicExtentWidget", "GeographicExtentMapWidget"]:
        prop.update(
            {
                "type": "array",
                "minItems": 4,
                "additionalItems": False,
                "items": [
                    {
                        "type": "number",
                        "maximum": 90,
                        "minimum": -90,
                        "description": "North border of the bounding box",
                    },
                    {
                        "type": "number",
                        "maximum": 180,
                        "minimum": -180,
                        "description": "West border of the bounding box",
                    },
                    {
                        "type": "number",
                        "maximum": 90,
                        "minimum": -90,
                        "description": "South border of the bounding box",
                    },
                    {
                        "type": "number",
                        "maximum": 180,
                        "minimum": -180,
                        "description": "East border of the bounding box",
                    },
                ],
            }
        )

    # DateRangeWidget is a calendar date picker
    if element["type"] == "DateRangeWidget":
        prop["description"] = "date formatted like yyyy-mm-dd/yyyy-mm-dd"

    # a single geographic location
    if element["type"] == "GeographicLocationWidget":
        prop.update(
            {
                "type": "object",
                "description": "Longitude and latitude of a single location",
                "properties": {
                    "longitude": {
                        "type": "number",
                        "maximum": 180,
                        "minimum": -180,
                    },
                    "latitude": {
                        "type": "number",
                        "maximum": 90,
                        "minimum": -90,
                    },
                },
            }
        )

    if description := element.get("help"):
        prop["description"] = description


def ecmwf_format(v: str, alias: bool = True) -> str:
    """Add ECMWF prefix to value v if v is a ECMWF keyword.

    :param v: parameter to format
    :param alias: whether to format for alias (with ':') or for query param (False, with '_')
    :return: formatted parameter

    >>> ecmwf_format('dataset', alias=False)
    'ecmwf_dataset'
    >>> ecmwf_format('variable')
    'ecmwf:variable'
    >>> ecmwf_format('unknown_param')
    'unknown_param'
    """
    separator = ":" if alias else "_"
    return f"{ECMWF_PREFIX[:-1]}{separator}{v}" if v in ALLOWED_KEYWORDS else v


def ecmwf_temporal_to_eodag(
    params: dict[str, Any],
) -> tuple[Optional[str], Optional[str]]:
    """
    Converts ECMWF temporal parameters to EODAG temporal parameters.

    ECMWF temporal parameters:
        - **year** or **hyear**: Union[str, list[str]] — Year(s) as a string or list of strings.
        - **month** or **hmonth**: Union[str, list[str]] — Month(s) as a string or list of strings.
        - **day** or **hday**: Union[str, list[str]] — Day(s) as a string or list of strings.
        - **time**: str — A string representing the time in the format `HHMM` (e.g., `0200`, `0800`, `1400`).
        - **date**: str — A string in one of the formats:
            - `YYYY-MM-DD`
            - `YYYY-MM-DD/YYYY-MM-DD`
            - `YYYY-MM-DD/to/YYYY-MM-DD`

    :param params: Dictionary containing ECMWF temporal parameters.
    :return: A tuple with:
        - **start**: A string in the format `YYYY-MM-DDTHH:MM:SSZ`.
        - **end**: A string in the format `YYYY-MM-DDTHH:MM:SSZ`.
    """
    start = end = None

    if date := params.get("date"):
        if isinstance(date, list):
            date = "/".join(date)
        start, end = parse_date(date, params.get("time"))

    elif year := (params.get("year") or params.get("hyear")):
        month = params.get("month") or params.get("hmonth")
        day = params.get("day") or params.get("hday")
        time = params.get("time")

        start, end = parse_year_month_day(year, month, day, time)

    if start and end:
        return to_iso_utc_string(start), to_iso_utc_string(end)
    else:
        return None, None


class ECMWFSearch(PostJsonSearch):
    """ECMWF search plugin.

    This plugin builds a :class:`~eodag.api.search_result.SearchResult` containing a single product
    using given query parameters as product properties.

    The available configuration parameters inherits from parent classes, with some particular parameters
    for this plugin.

    :param provider: An eodag providers configuration dictionary
    :param config: Search plugin configuration:

        * :attr:`~eodag.config.PluginConfig.remove_from_query` (``list[str]``): List of parameters
          used to parse metadata but that must not be included to the query
        * :attr:`~eodag.config.PluginConfig.end_date_excluded` (``bool``): Set to `False` if
          provider does not include end date to search
        * :attr:`~eodag.config.PluginConfig.discover_queryables`
          (:class:`~eodag.config.PluginConfig.DiscoverQueryables`): configuration to fetch the queryables from a
          provider queryables endpoint; It has the following keys:

          * :attr:`~eodag.config.PluginConfig.DiscoverQueryables.fetch_url` (``str``): url to fetch the queryables valid
            for all collections
          * :attr:`~eodag.config.PluginConfig.DiscoverQueryables.collection_fetch_url` (``str``): url to fetch the
            queryables for a specific collection
          * :attr:`~eodag.config.PluginConfig.DiscoverQueryables.constraints_url` (``str``): url of the constraint file
            used to build queryables

        * :attr:`~eodag.config.PluginConfig.dynamic_discover_queryables`
          (``list`` [:class:`~eodag.config.PluginConfig.DynamicDiscoverQueryables`]): list of configurations to fetch
          the queryables from different provider queryables endpoints. A configuration is used based on the given
          selection criterias. The first match is used. If no match is found, it falls back to standard behaviors
          (e.g. discovery using :attr:`~eodag.config.PluginConfig.discover_queryables`).
          Each element of the list has the following keys:

          * :attr:`~eodag.config.PluginConfig.DynamicDiscoverQueryables.collection_selector`
            (``list`` [:class:`~eodag.config.PluginConfig.CollectionSelector`]): list of collection selection
            criterias. The configuration given in
            :attr:`~eodag.config.PluginConfig.DynamicDiscoverQueryables.discover_queryables` is used if any collection
            selector matches the search parameters. The selector matches if the field value starts with the given
            prefix, i.e. it matches if ``parameters[field].startswith(prefix)==True``. It has the following keys:

            * :attr:`~eodag.config.PluginConfig.CollectionSelector.field` (``str``) Field in the search parameters to
              match
            * :attr:`~eodag.config.PluginConfig.CollectionSelector.prefix` (``str``) Prefix to match in the field

          * :attr:`~eodag.config.PluginConfig.DynamicDiscoverQueryables.discover_queryables`
            (``list`` [:class:`~eodag.config.PluginConfig.DiscoverQueryables`]): same as
            :attr:`~eodag.config.PluginConfig.discover_queryables` above.
    """

    def __init__(self, provider: str, config: PluginConfig) -> None:
        config.metadata_mapping = {
            **ecmwf_mtd(),
            **{
                "id": "$.id",
                "title": "$.id",
                "order:status": OFFLINE_STATUS,
                "eodag:download_link": "$.null",
                "geometry": ["feature", "$.geometry"],
                "eodag:default_geometry": "POLYGON((180 -90, 180 90, -180 90, -180 -90, 180 -90))",
            },
            **config.metadata_mapping,
        }

        super().__init__(provider, config)

        # ECMWF providers do not feature any api_endpoint or next_page_query_obj.
        # Searched is faked by EODAG.
        self.config.__dict__.setdefault("api_endpoint", "")
        self.config.pagination.setdefault("next_page_query_obj", "{{}}")

        # defaut conf for accepting custom query params
        self.config.__dict__.setdefault(
            "discover_metadata",
            {
                "auto_discovery": False,
                "search_param": "{metadata}",
                "metadata_pattern": "^[a-zA-Z0-9][a-zA-Z0-9_]*$",
            },
        )

    def do_search(
        self, prep: PreparedSearch = PreparedSearch(limit=None), **kwargs: Any
    ) -> RawSearchResult:
        """Should perform the actual search request.

        :param args: arguments to be used in the search
        :param kwargs: keyword arguments to be used in the search
        :return: list containing the results from the provider in json format
        """
        # no real search. We fake it all
        raw_search_results = RawSearchResult([{}])
        raw_search_results.search_params = kwargs
        raw_search_results.query_params = (
            prep.query_params if hasattr(prep, "query_params") else {}
        )
        raw_search_results.collection_def_params = (
            prep.collection_def_params if hasattr(prep, "collection_def_params") else {}
        )
        return raw_search_results

    def query(
        self,
        prep: PreparedSearch = PreparedSearch(),
        **kwargs: Any,
    ) -> SearchResult:
        """Build ready-to-download SearchResult

        :param prep: :class:`~eodag.plugins.search.PreparedSearch` object containing information needed for the search
        :param kwargs: keyword arguments to be used in the search
        :returns: list of products and number of products (optional)
        """
        collection = prep.collection
        if not collection:
            collection = kwargs.get("collection")
        kwargs = self._preprocess_search_params(kwargs)
        result = super().query(prep, **kwargs)
        if prep.count and not result.number_matched:
            result.number_matched = 1

        return result

    def clear(self) -> None:
        """Clear search context"""
        super().clear()

    def build_query_string(
        self, collection: str, query_dict: dict[str, Any]
    ) -> tuple[dict[str, Any], str]:
        """Build The query string using the search parameters

        :param collection: collection id
        :param query_dict: keyword arguments to be used in the query string
        :return: formatted query params and encode query string
        """
        query_dict["_date"] = f"{query_dict.get(START)}/{query_dict.get(END)}"

        # Reorder kwargs to make sure year/month/day/time if set overwrite default datetime.
        priority_keys = [
            START,
            END,
        ]
        ordered_kwargs = {k: query_dict[k] for k in priority_keys if k in query_dict}
        ordered_kwargs.update(query_dict)

        return super().build_query_string(
            collection=collection, query_dict=ordered_kwargs
        )

    def _preprocess_search_params(self, params: dict[str, Any]) -> dict[str, Any]:
        """Preprocess search parameters before making a request to the CDS API.

        This method is responsible for checking and updating the provided search parameters
        to ensure that required parameters like 'collection', 'start_datetime',
        'end_datetime', and 'geometry' are properly set. If not specified
        in the input parameters, default values or values from the configuration are used.

        :param params: Search parameters to be preprocessed.
        """

        _dc_qs = params.get("_dc_qs")

        if _dc_qs is not None:
            # if available, update search params using datacube query-string
            _dc_qp = geojson.loads(unquote_plus(unquote_plus(_dc_qs)))
            if "/to/" in _dc_qp.get("date", ""):
                params[START], params[END] = _dc_qp["date"].split("/to/")
            elif "/" in _dc_qp.get("date", ""):
                (params[START], params[END],) = _dc_qp[
                    "date"
                ].split("/")
            elif _dc_qp.get("date"):
                params[START] = params[END] = _dc_qp["date"]

            if "/" in _dc_qp.get("area", ""):
                params["geometry"] = _dc_qp["area"].split("/")

        params = {
            k.removeprefix(ECMWF_PREFIX).removeprefix(f"{ECMWF_PREFIX[:-1]}_"): v
            for k, v in params.items()
            if v is not None
        }

        # read 'start_datetime' and 'end_datetime' from 'date' range
        if "date" in params:
            start_date, end_date = parse_date(params["date"])
            params[START] = format_date(start_date)
            params[END] = format_date(end_date)

        # adapt end date if it is midnight
        if END in params:
            end_date_excluded = getattr(self.config, "end_date_excluded", True)
            is_datetime = "T" in str(params[END])
            end_date = parse_to_utc(params[END])
            start_date = parse_to_utc(params[START])
            if (
                not end_date_excluded
                and is_datetime
                and end_date > start_date
                and end_date
                == end_date.replace(hour=0, minute=0, second=0, microsecond=0)
            ):
                end_date += dt.timedelta(days=-1)
                params[END] = to_iso_utc_string(end_date)

        # geometry
        if "geometry" in params:
            params["geometry"] = get_geometry_from_various(geometry=params["geometry"])

        # check ecmwf geom format if given
        # ECMWF Polytope uses non-geojson structure for features
        if "feature" in params:
            get_geometry_from_ecmwf_feature(params["feature"])
        # bounding box in area format
        if "area" in params:
            get_geometry_from_ecmwf_area(params["area"])
        # single location
        if "location" in params:
            get_geometry_from_ecmwf_location(params["location"])

        return params

    def _preprocess_indirect_date_parameters(self, params: dict[str, Any]) -> dict:
        """
        Compute start_datetime / end_datetime from "date", "time", "year", "month", "day"
        """
        indirects: dict[str, Any] = {}

        # Validate and collect indirect date parameters
        time = validate_datetime_param(
            params.get("time"), "time", ["%H%M", "%H:%M", "%H%M%S", "%H:%M:%S"]
        )
        year = validate_datetime_param(params.get("year"), "year", ["%Y"])
        month = validate_datetime_param(params.get("month"), "month", ["%m"])
        day = validate_datetime_param(params.get("day"), "day", ["%d"])

        if time is not None:
            time = time_values_to_hhmm(time)
            indirects["time"] = time
        if year is not None:
            indirects["year"] = year
        if month is not None:
            indirects["month"] = [f"{int(m):02d}" for m in month]
        if day is not None:
            indirects["day"] = [f"{int(d):02d}" for d in day]

        # Compute date range from "date" param (takes precedence)
        date = params.get("date", None)
        if date is not None:
            try:
                start, end = compute_date_range_from_params(date=date, time=time)
                indirects["start_datetime"] = start
                indirects["end_datetime"] = end
                return indirects
            except Exception as e:
                raise ValidationError(
                    'Malformed parameter "date" (date given "{}", time given: "{}"): {}'.format(
                        params.get("date"), params.get("time"), str(e)
                    )
                )

        # Compute date range from year/month/day/time params
        start, end = compute_date_range_from_params(
            year=year,
            month=indirects.get("month"),
            day=indirects.get("day"),
            time=time,
        )
        if start is not None:
            indirects["start_datetime"] = start
        if end is not None:
            indirects["end_datetime"] = end

        return indirects

    def _get_collection_queryables(
        self, collection: Optional[str], alias: Optional[str], filters: dict[str, Any]
    ) -> QueryablesDict:
        """Override to set additional_properties to false."""
        default_values: dict[str, Any] = deepcopy(
            getattr(self.config, "products", {}).get(collection, {})
        )
        default_values.pop("metadata_mapping", None)
        default_values.pop("metadata_mapping_from_product", None)

        filters["collection"] = collection
        queryables = self.discover_queryables(**{**default_values, **filters}) or {}

        return QueryablesDict(additional_properties=False, **queryables)

    def _find_dynamic_queryables_config(
        self, kwargs: dict[str, Any], dynamic_config: list
    ) -> dict[str, Any]:
        """Find the appropriate queryables configuration from dynamic configuration.

        :param kwargs: Search parameters
        :param dynamic_config: List of dynamic discover queryables configurations
        :return: Found queryables configuration or empty dict
        """
        for dc in dynamic_config:
            for cs in dc["collection_selector"]:
                field = cs["field"]
                if kwargs[field].startswith(cs["prefix"]):
                    return dc["discover_queryables"]
        return {}

    def _is_discoverable_metadata_key(self, key: str) -> bool:
        """Check if a key can bypass strict queryables validation via discover_metadata."""

        discover_metadata = getattr(self.config, "discover_metadata", None) or {}
        if not discover_metadata.get("auto_discovery"):
            return False

        pattern = discover_metadata.get("metadata_pattern") or ""
        try:
            return bool(pattern and re.match(pattern, key))
        except re.error:
            msg = f"Invalid discover_metadata.metadata_pattern for provider {self.provider}: {pattern}"
            raise MisconfiguredError(msg)

    def discover_queryables(
        self,
        **kwargs: Any,
    ) -> Optional[dict[str, Annotated[Any, FieldInfo]]]:
        """Fetch queryables list from provider using its constraints file

        :param kwargs: additional filters for queryables (`collection` and other search
                       arguments)
        :returns: fetched queryable parameters dict
        """
        collection = kwargs.pop("collection")

        col_config = self.get_collection_def_params(collection)

        default_values = deepcopy(col_config)
        default_values.pop("metadata_mapping", None)
        default_values.pop("metadata_mapping_from_product", None)
        default_values.pop("discover_queryables", None)
        kwargs.pop("discover_queryables", None)
        filters = {**default_values, **kwargs}

        if "start" in filters:
            filters[START] = filters.pop("start")
        if "end" in filters:
            filters[END] = filters.pop("end")

        # extract default datetime and convert geometry
        try:
            processed_filters = self._preprocess_search_params(deepcopy(filters))
        except Exception as e:
            raise ValidationError(e.args[0]) from e

        # dynamic_discover_queryables for WekeoECMWFSearch
        queryables_config = {}
        if dynamic_config := getattr(self.config, "dynamic_discover_queryables", []):
            queryables_config = self._find_dynamic_queryables_config(
                kwargs, dynamic_config
            )

        provider_dq = getattr(self.config, "discover_queryables", {}) or {}
        product_dq = col_config.get("discover_queryables", {}) or {}
        dq_conf = {**provider_dq, **product_dq, **queryables_config}
        constraints_url = format_metadata(dq_conf.get("constraints_url", ""), **filters)
        constraints: list[dict[str, Any]] = self._fetch_data(constraints_url)

        form_url = format_metadata(dq_conf.get("form_url", ""), **filters)
        form: list[dict[str, Any]] = self._fetch_data(form_url)

        formated_filters = self.format_as_provider_keyword(
            collection, deepcopy(processed_filters)
        )
        # we re-apply kwargs input to consider override of year, month, day and time.
        for k, v in {**default_values, **kwargs}.items():
            key = k.removeprefix(ECMWF_PREFIX).removeprefix(f"{ECMWF_PREFIX[:-1]}_")

            if key not in ALLOWED_KEYWORDS | {
                START,
                END,
                "geometry",
            } and not self._is_discoverable_metadata_key(key):
                raise ValidationError(
                    f"'{key}' is not a queryable parameter for {self.provider}", {key}
                )

            formated_filters[key] = v

        # we use non empty filters as default to integrate user inputs
        # it is needed because pydantic json schema does not represent "value"
        # but only "default"
        non_empty_formated: dict[str, Any] = {
            k: v
            for k, v in formated_filters.items()
            if v and (not isinstance(v, list) or all(v))
        }

        required_keywords: set[str] = set()

        # calculate available values
        if constraints:
            # Apply constraints filtering
            available_values = self.available_values_from_constraints(
                constraints,
                non_empty_formated,
                form_keywords=[f["name"] for f in form],
            )

            # Pre-compute the required keywords (present in all constraint dicts)
            # when form, required keywords are extracted directly from form
            if not form:
                required_keywords = set.intersection(
                    *(map(lambda d: set(d.keys()), constraints))
                )

        else:
            values_url = getattr(self.config, "available_values_url", "")
            if not values_url:
                return self.queryables_from_metadata_mapping(collection)
            if "{" in values_url:
                values_url = format_string(None, values_url, **filters)
            data = self._fetch_data(values_url)
            available_values = data["constraints"]
            required_keywords = data.get("required", [])

        # To check if all keywords are queryable parameters, we check if they are in the
        # available values or the collection config (available values calculated from the
        # constraints might not include all queryables)
        for keyword in processed_filters:
            if (
                keyword
                not in available_values.keys()
                | col_config.keys()
                | {
                    START,
                    END,
                    "geometry",
                }
                and keyword not in [f["name"] for f in form]
                and keyword
                not in set(list(available_values.keys()) + [f["name"] for f in form])
                and not self._is_discoverable_metadata_key(keyword)
            ):
                raise ValidationError("'%s' is not a queryable parameter" % keyword)

        # generate queryables
        if form:
            queryables = self.queryables_by_form(
                form,
                available_values,
                non_empty_formated,
            )
        else:
            queryables = self.queryables_by_values(
                available_values, list(required_keywords), non_empty_formated
            )

        # ecmwf:date is replaced by start and end.
        # start and end filters are supported whenever combinations of "year", "month", "day" filters exist
        queryable_prefix = f"{ECMWF_PREFIX[:-1]}_"
        if (
            f"{queryable_prefix}date" in queryables
            or f"{queryable_prefix}year" in queryables
            or f"{queryable_prefix}hyear" in queryables
        ):
            queryables.update(
                {
                    "start": Queryables.get_with_default(
                        "start", processed_filters.get(START)
                    ),
                    "end": Queryables.get_with_default(
                        "end",
                        processed_filters.get(END),
                    ),
                }
            )

        # area is geom in EODAG.
        if queryables.pop("area", None):
            queryables["geom"] = Queryables.get_with_default("geom", None)

        return queryables

    def available_values_from_constraints(
        self,
        constraints: list[dict[str, Any]],
        input_keywords: dict[str, Any],
        form_keywords: list[str],
    ) -> dict[str, list[str]]:
        """
        Filter constraints using input_keywords. Return list of available queryables.
        All constraint entries must have the same parameters.

        :param constraints: list of constraints received from the provider
        :param input_keywords: dict of input parameters given by the user
        :param form_keywords: list of keyword names from the provider form endpoint
        :return: dict with available values for each parameter
        """
        # get ordered constraint keywords
        constraints_keywords = list(
            OrderedDict.fromkeys(k for c in constraints for k in c.keys())
        )

        # prepare ordered input keywords formatted as provider's keywords
        # required to filter with constraints
        ordered_keywords = (
            [kw for kw in form_keywords if kw in constraints_keywords]
            if form_keywords
            else constraints_keywords
        )

        # filter constraint entries matching input keyword values
        filtered_constraints: list[dict[str, Any]]

        parsed_keywords: list[str] = []
        for keyword in ordered_keywords:
            values = input_keywords.get(keyword)

            if values is None:
                parsed_keywords.append(keyword)
                continue

            # we only compare list of strings.
            if isinstance(values, dict):
                raise ValidationError(
                    f"Parameter value as object is not supported: {keyword}={values}",
                    {keyword},
                )

            # We convert every single value to a list of string
            filter_v = list(values) if isinstance(values, tuple) else values
            filter_v = filter_v if isinstance(filter_v, list) else [filter_v]
            filter_v = [str(v) for v in filter_v]

            # We strip values of superfluous quotes (added by mapping converter to_geojson).
            # ECMWF accept date ranges with /to/. We need to split it to an array
            # ECMWF accept date ranges in format val1/val2. We need to split it to an array
            date_regex = [
                re.compile(p) for p in (DATE_RANGE_PATTERN, COMPACT_DATE_RANGE_PATTERN)
            ]
            is_date = any(
                any(r.match(v) is not None for r in date_regex) for v in filter_v
            )
            if is_date:
                sep = re.compile(r"/to/|/")
                filter_v = [i for v in filter_v for i in sep.split(str(v))]

            # special handling for time 0000 converted to 0 by pre-formating with metadata_mapping
            if keyword.split(":")[-1] == "time":
                filter_v = ["0000" if str(v) == "0" else v for v in filter_v]

            # Collect missing values to report errors
            missing_values = set(filter_v)

            # Filter constraints and check for missing values
            filtered_constraints = []
            # True if some constraint is defined for this keyword.
            # In other words: if no constraint defines a list of values
            # then any value is allowed for this keyword
            keyword_constrained = False
            for entry in constraints:
                # Filter based on the presence of any value in filter_v
                entry_values = entry.get(keyword, [])
                if entry_values:
                    keyword_constrained = True

                # date constraint may be intervals. We identify intervals with a "/" in the value.
                # date constraint can be a mixed list of single values (e.g "2023-06-27")
                # and intervals (e.g. "2024-11-12/2025-11-20").
                # collections with mixed values: CAMS_GAC_FORECAST, CAMS_EU_AIR_QUALITY_FORECAST
                present_values = []
                for entry_value in entry_values:
                    if keyword == "date" and "/" in entry_value:
                        input_range = values
                        if isinstance(values, list):
                            input_range = values[0]
                        if "/" not in input_range:
                            input_range = f"{input_range}/{input_range}"
                        if is_range_in_range(entry_value, input_range):
                            present_values.extend(filter_v)
                    else:
                        new_values = [
                            value for value in filter_v if value == entry_value
                        ]
                        present_values.extend(new_values)

                # Remove present values from the missing_values set
                missing_values -= set(present_values)

                if present_values:
                    filtered_constraints.append(entry)

            # raise an error as no constraint entry matched the input keywords
            # raise an error if one value from input is not allowed
            if keyword_constrained and (not filtered_constraints or missing_values):
                allowed_values = list(
                    {value for c in constraints for value in c.get(keyword, [])}
                )
                # restore ecmwf: prefix before raising error
                keyword = ECMWF_PREFIX + keyword

                all_keywords_str = ""
                if len(parsed_keywords) > 1:
                    keywords = [
                        f"{ECMWF_PREFIX + k}={pk}"
                        for k in parsed_keywords
                        if (pk := input_keywords.get(k))
                    ]
                    all_keywords_str = f" with {', '.join(keywords)}"

                raise ValidationError(
                    f"{keyword}={values} is not available"
                    f"{all_keywords_str}."
                    f" Allowed values are {', '.join(allowed_values)}.",
                    set(
                        [keyword] + [k for k in parsed_keywords if k in input_keywords]
                    ),
                )

            parsed_keywords.append(keyword)
            # if the keyword is not constrained then any value is allowed
            if keyword_constrained:
                constraints = filtered_constraints

        available_values: dict[str, Any] = {k: set() for k in ordered_keywords}

        # we aggregate the constraint entries left
        for entry in constraints:
            for key, value in entry.items():
                available_values[key].update(value)

        return {k: list(v) for k, v in available_values.items()}

    def queryables_by_form(
        self,
        form: list[dict[str, Any]],
        available_values: dict[str, list[str]],
        defaults: dict[str, Any],
    ) -> dict[str, Annotated[Any, FieldInfo]]:
        """
        Generate Annotated field definitions from form entries and available values
        Used by Copernicus services like cop_cds, cop_ads, cop_ewds.

        :param form: data fetched from the form endpoint of the provider
        :param available_values: available values for each parameter
        :param defaults: default values for the parameters
        :return: dict of annotated queryables
        """
        queryables: dict[str, Annotated[Any, FieldInfo]] = {}

        required_list: list[str] = []
        for element in form:
            name: str = element["name"]

            # those are not parameter elements.
            if name in ("area_group", "global", "warning", "licences"):
                continue
            if "type" not in element or element["type"] == "FreeEditionWidget":
                # FreeEditionWidget used to select the whole available region
                # and to provide comments for the dataset
                continue

            # ordering done by id -> set id to high value if not present -> element will be last
            if "id" not in element:
                element["id"] = 100

            prop = {"title": element.get("label", name)}

            details = element.get("details", {})

            # add values from form if keyword was not in constraints
            values = (
                available_values[name]
                if name in available_values
                else details.get("values")
            )

            # updates the properties with the values given based on the information from the element
            _update_properties_from_element(prop, element, values)

            default = defaults.get(name)

            if details:
                fields = details.get("fields")
                if fields and (comment := fields[0].get("comment")):
                    prop["description"] = comment

            if name == "area" and isinstance(default, dict):
                default = list(default.values())

            # sometimes form returns default as array instead of string
            if default and prop.get("type") == "string" and isinstance(default, list):
                default = ",".join(default)

            is_required: bool
            if available_values.get(name):
                # required by the filtered constraints (available_values[name] is a not empty list)
                is_required = True
            elif bool(element.get("required")):
                if name in available_values and not available_values[name]:
                    # not required by the filtered constraints (available_values[name] is an empty list)
                    is_required = False
                else:
                    # required only by form
                    is_required = True
            else:
                # not required by form
                is_required = False

            if is_required:
                required_list.append(name)

            formatted_param = ecmwf_format(name, alias=False)
            formatted_alias = ecmwf_format(name)
            queryables[formatted_param] = Annotated[
                get_args(
                    json_field_definition_to_python(
                        prop,
                        default_value=default,
                        required=is_required,
                        validation_alias=AliasChoices(formatted_alias, name),
                        serialization_alias=formatted_alias,
                    )
                )
            ]

        return queryables

    def queryables_by_values(
        self,
        available_values: dict[str, list[str]],
        required_keywords: list[str],
        defaults: dict[str, Any],
    ) -> dict[str, Annotated[Any, FieldInfo]]:
        """
        Generate Annotated field definitions from available values.
        Used by ECMWF data providers like dedt_lumi.

        :param available_values: available values for each parameter
        :param required_keywords: list of required parameters
        :param defaults: default values for the parameters
        :return: dict of annotated queryables
        """
        # Rename keywords from form with metadata mapping.
        # Needed to map constraints like "xxxx" to eodag parameter "ecmwf:xxxx"
        required = [ecmwf_format(k) for k in required_keywords]  # noqa: F841

        queryables: dict[str, Annotated[Any, FieldInfo]] = {}
        for name, values in available_values.items():
            # Rename keywords from form with metadata mapping.
            # Needed to map constraints like "xxxx" to eodag parameter "ecmwf:xxxx"
            formatted_param = ecmwf_format(name, alias=False)
            formatted_alias = ecmwf_format(name)

            queryables[formatted_param] = Annotated[
                get_args(
                    json_field_definition_to_python(
                        {"type": "string", "title": name, "enum": values},
                        default_value=defaults.get(name),
                        required=bool(formatted_alias in required),
                        validation_alias=AliasChoices(formatted_alias, name),
                        serialization_alias=formatted_alias,
                    )
                )
            ]

        return queryables

    def format_as_provider_keyword(
        self, collection: str, properties: dict[str, Any]
    ) -> dict[str, Any]:
        """Return provider equivalent keyword names from EODAG keywords.

        :param collection: collection id
        :param properties: dict of properties to be formatted
        :return: dict of formatted properties
        """
        properties["collection"] = collection

        # provider collection specific conf
        collection_def_params = self.get_collection_def_params(
            collection, format_variables=properties
        )

        # Add to the query, the queryable parameters set in the provider collection definition
        properties.update(
            {
                k: v
                for k, v in collection_def_params.items()
                if k not in properties.keys()
                and k in self.config.metadata_mapping.keys()
                and isinstance(self.config.metadata_mapping[k], list)
            }
        )
        qp, _ = self.build_query_string(collection, properties)

        return qp

    @instance_cached_method()
    def _fetch_data(self, url: str) -> Any:
        """
        fetches from a provider elements like constraints or forms.

        :param url: url from which the constraints can be fetched
        :returns: json file content fetched from the provider
        """
        if not url:
            return []

        auth = (
            self.auth
            if hasattr(self, "auth") and isinstance(self.auth, AuthBase)
            else None
        )
        timeout = getattr(self.config, "timeout", DEFAULT_SEARCH_TIMEOUT)
        return fetch_json(url, auth=auth, timeout=timeout)

    def normalize_results(
        self, results: RawSearchResult, **kwargs: Any
    ) -> list[EOProduct]:
        """Build :class:`~eodag.api.product._product.EOProduct` from provider result

        :param results: Raw provider result as single dict in list
        :param kwargs: Search arguments
        :returns: list of single :class:`~eodag.api.product._product.EOProduct`
        """

        collection = kwargs.get("collection")

        result = results[0]

        # datacube query string got from previous search
        _dc_qs = kwargs.pop("_dc_qs", None)
        if _dc_qs is not None:
            qs = unquote_plus(unquote_plus(_dc_qs))
            sorted_unpaginated_qp = geojson.loads(qs)
        else:
            sorted_unpaginated_qp = dict_items_recursive_sort(results.query_params)

        # remove unwanted query params
        for param in getattr(self.config, "remove_from_query", []):
            sorted_unpaginated_qp.pop(param, None)

        if result:
            properties = result
            properties.update(result.pop("eodag:request_params", None) or {})

            properties = {k: v for k, v in properties.items() if not k.startswith("__")}

            properties["geometry"] = properties.get("area") or DEFAULT_GEOMETRY

            start, end = ecmwf_temporal_to_eodag(properties)
            properties["start_datetime"] = start
            properties["end_datetime"] = end

        else:
            # use all available query_params to parse properties
            result_data: dict[str, Any] = {
                **results.collection_def_params,
                **sorted_unpaginated_qp,
                **{"qs": sorted_unpaginated_qp},
            }

            # update result with collection_def_params and search args if not None (and not auth)
            kwargs.pop("auth", None)
            result_data.update(results.collection_def_params)
            result_data = {
                **result_data,
                **{k: v for k, v in kwargs.items() if v is not None},
            }

            # start_datetime /  computed from "date", "time", "year", "month", "day"
            indirects = self._preprocess_indirect_date_parameters(result_data)
            for key in (START, END):
                if result_data.get(key) is None and indirects.get(key) is not None:
                    result_data[key] = indirects[key]

            properties = properties_from_json(
                result_data,
                self.config.metadata_mapping,
                discovery_config=getattr(self.config, "discover_metadata", {}),
            )

            query_hash = hashlib.sha1(str(result_data).encode("UTF-8")).hexdigest()

            properties["title"] = properties["id"] = (
                (collection or kwargs.get("dataset", self.provider)).upper()
                + "_ORDERABLE_"
                + query_hash
            )

            # collection alias (required by opentelemetry-instrumentation-eodag)
            if alias := getattr(self.config, "collection_config", {}).get("alias"):
                kwargs["collection"] = alias

        # Convert ecmwf geometries for properties but keep original in qs
        # ECMWF Polytope uses non-geojson structure for features
        if "feature" in sorted_unpaginated_qp:
            properties["geometry"] = get_geometry_from_ecmwf_feature(
                sorted_unpaginated_qp["feature"]
            )
            properties.pop("feature", None)
        # bounding box in area format
        if "area" in sorted_unpaginated_qp:
            properties["geometry"] = get_geometry_from_ecmwf_area(
                sorted_unpaginated_qp["area"]
            )
            properties.pop("area", None)
        # single location
        if "location" in sorted_unpaginated_qp:
            properties["geometry"] = get_geometry_from_ecmwf_location(
                sorted_unpaginated_qp["location"]
            )
            properties.pop("location", None)

        qs = geojson.dumps(sorted_unpaginated_qp)

        # used by server mode to generate eodag:download_link href
        # TODO: to remove once the legacy server is removed
        properties["_dc_qs"] = quote_plus(qs)

        product = EOProduct(
            provider=self.provider,
            properties={ecmwf_format(k): v for k, v in properties.items()},
            **kwargs,
        )

        # backup original register_downloader to register_downloader_only
        product.register_downloader_only = product.register_downloader
        # patched register_downloader that will also update properties
        product.register_downloader = MethodType(patched_register_downloader, product)  # type: ignore[method-assign]

        return [product]

    def count_hits(
        self, count_url: Optional[str] = None, result_type: Optional[str] = None
    ) -> int:
        """Count method that will always return 1.

        :param count_url: not used, only here because this method overwrites count_hits from the parent class
        :param result_type: not used, only here because this method overwrites count_hits from the parent class
        :return: always 1
        """
        return 1


def _check_id(product: EOProduct) -> EOProduct:
    """Check if the id is the one of an existing job.

    If the job exists, poll it, otherwise, raise an error.

    :param product: The product to check the id for
    :raises: :class:`~eodag.utils.exceptions.ValidationError`
    """
    if not (product_id := product.search_kwargs.get("id")):
        return product

    if "ORDERABLE" in product_id:
        return product

    if product.downloader is None:
        return product
    on_response_mm = getattr(product.downloader.config, "order_on_response", {}).get(
        "metadata_mapping", {}
    )
    if not on_response_mm:
        return product

    logger.debug(f"Update product properties using given eodag:order_id {product_id}")
    on_response_mm_jsonpath = mtd_cfg_as_conversion_and_querypath(
        on_response_mm,
    )
    properties_update = properties_from_json(
        {}, {**on_response_mm_jsonpath, **{"eodag:order_id": (None, product_id)}}
    )
    product.properties.update(
        {k: v for k, v in properties_update.items() if v != NOT_AVAILABLE}
    )

    auth = product.downloader_auth.authenticate() if product.downloader_auth else None

    # try to poll the job corresponding to the given id
    try:
        product.downloader._order_status(product=product, auth=auth)  # type: ignore
    # when a NotAvailableError is catched, it means the product is not ready and still needs to be polled
    except NotAvailableError:
        product.properties["order:status"] = STAGING_STATUS
    except Exception as e:
        if (
            isinstance(e, DownloadError) or isinstance(e, ValidationError)
        ) and "order status could not be checked" in e.args[0]:
            raise ValidationError(
                f"Requested data is not available on {product.provider} ({product_id})."
            ) from e
        raise ValidationError(e.args[0]) from e

    # update product id
    product.properties["id"] = product_id
    # update collection if needed
    if product.collection is None:
        product.collection = product.properties.get("ecmwf:dataset")
    # update product title
    product.properties["title"] = (
        (product.collection or product.provider).upper() + "_" + product_id
    )
    # use NOT_AVAILABLE as fallback collection to avoid using guess_collection
    if product.collection is None:
        product.collection = NOT_AVAILABLE

    return product


def patched_register_downloader(self, downloader, authenticator):
    """Register product donwloader and update properties if searched by id.

    :param self: product to which information should be added
    :param downloader: The download method that it can use
                    :class:`~eodag.plugins.download.base.Download` or
                    :class:`~eodag.plugins.api.base.Api`
    :param authenticator: The authentication method needed to perform the download
                        :class:`~eodag.plugins.authentication.base.Authentication`
    """
    # register downloader
    self.register_downloader_only(downloader, authenticator)
    # and also update properties
    _check_id(self)


class MeteoblueSearch(ECMWFSearch):
    """MeteoblueSearch search plugin.

    This plugin, which inherits from :class:`~eodag.plugins.search.build_search_result.ECMWFSearch`,
    performs a POST request and uses its result to build a single :class:`~eodag.api.search_result.SearchResult`
    object.

    The available configuration parameters are inherited from parent classes, with some a particularity
    for pagination for this plugin.

    :param provider: An eodag providers configuration dictionary
    :param config: Search plugin configuration:

        * :attr:`~eodag.config.PluginConfig.pagination` (:class:`~eodag.config.PluginConfig.Pagination`)
          (**mandatory**): The configuration of how the pagination is done on the provider. For
          this plugin it has the node:

          * :attr:`~eodag.config.PluginConfig.Pagination.next_page_query_obj` (``str``): The
            additional parameters needed to perform search. These parameters won't be included in
            the result. This must be a json dict formatted like ``{{"foo":"bar"}}`` because it
            will be passed to a :meth:`str.format` method before being loaded as json.
    """

    def collect_search_urls(
        self,
        prep: PreparedSearch = PreparedSearch(),
        **kwargs: Any,
    ) -> tuple[list[str], int]:
        """Wraps PostJsonSearch.collect_search_urls to force product count to 1

        :param prep: :class:`~eodag.plugins.search.PreparedSearch` object containing information for the search
        :param kwargs: keyword arguments used in the search
        :return: list of search url and number of results
        """
        urls, _ = super().collect_search_urls(prep, **kwargs)
        return urls, 1

    def do_search(
        self, prep: PreparedSearch = PreparedSearch(limit=None), **kwargs: Any
    ) -> RawSearchResult:
        """Perform the actual search request, and return result in a single element.

        :param prep: :class:`~eodag.plugins.search.PreparedSearch` object containing information for the search
        :param kwargs: keyword arguments to be used in the search
        :return: list containing the results from the provider in json format
        """

        prep.url = prep.search_urls[0]
        prep.info_message = f"Sending search request: {prep.url}"
        prep.exception_message = (
            f"Skipping error while searching for {self.provider}"
            f" {self.__class__.__name__} instance"
        )
        response = self._request(prep)
        raw_search_results = RawSearchResult([response.json()])
        raw_search_results.search_params = kwargs

        raw_search_results.query_params = prep.query_params
        raw_search_results.collection_def_params = prep.collection_def_params
        return raw_search_results

    def build_query_string(
        self, collection: str, query_dict: dict[str, Any]
    ) -> tuple[dict[str, Any], str]:
        """Build The query string using the search parameters

        :param collection: collection id
        :param query_dict: keyword arguments to be used in the query string
        :return: formatted query params and encode query string
        """
        return QueryStringSearch.build_query_string(self, collection, query_dict)

    def normalize_results(self, results, **kwargs):
        """Build :class:`~eodag.api.product._product.EOProduct` from provider result

        :param results: Raw provider result as single dict in list
        :param kwargs: Search arguments
        :returns: list of single :class:`~eodag.api.product._product.EOProduct`
        """

        collection = kwargs.get("collection")

        result = results[0]

        # datacube query string got from previous search
        _dc_qs = kwargs.pop("_dc_qs", None)
        if _dc_qs is not None:
            qs = unquote_plus(unquote_plus(_dc_qs))
            sorted_unpaginated_query_params = geojson.loads(qs)
        else:
            next_page_query_obj = orjson.loads(
                self.config.pagination["next_page_query_obj"].format()
            )
            unpaginated_query_params = {
                k: v
                for k, v in results.query_params.items()
                if (k, v) not in next_page_query_obj.items()
            }
            # query hash, will be used to build a product id
            sorted_unpaginated_query_params = dict_items_recursive_sort(
                unpaginated_query_params
            )

        # use all available query_params to parse properties
        result = dict(
            result,
            **sorted_unpaginated_query_params,
            qs=sorted_unpaginated_query_params,
        )

        qs = geojson.dumps(sorted_unpaginated_query_params)

        query_hash = hashlib.sha1(str(qs).encode("UTF-8")).hexdigest()

        # update result with collection_def_params and search args if not None (and not auth)
        kwargs.pop("auth", None)
        result.update(results.collection_def_params)
        result = dict(result, **{k: v for k, v in kwargs.items() if v is not None})

        # parse properties
        parsed_properties = properties_from_json(
            result,
            self.config.metadata_mapping,
            discovery_config=getattr(self.config, "discover_metadata", {}),
        )

        properties = {ecmwf_format(k): v for k, v in parsed_properties.items()}
        # collection alias (required by opentelemetry-instrumentation-eodag)
        if alias := getattr(self.config, "collection_config", {}).get("alias"):
            collection = alias

        def slugify(date_str: str) -> str:
            return date_str.split("T")[0].replace("-", "")

        # build product id
        product_id = (collection or self.provider).upper()

        start = properties.get(START, NOT_AVAILABLE)
        end = properties.get(END, NOT_AVAILABLE)

        if start != NOT_AVAILABLE:
            product_id += f"_{slugify(start)}"
            if end != NOT_AVAILABLE:
                product_id += f"_{slugify(end)}"

        product_id += f"_{query_hash}"

        properties["id"] = properties["title"] = product_id

        # used by server mode to generate eodag:download_link href
        properties["_dc_qs"] = quote_plus(qs)

        product = EOProduct(
            provider=self.provider,
            collection=collection,
            properties=properties,
        )

        return [
            product,
        ]


class WekeoECMWFSearch(ECMWFSearch):
    """
    WekeoECMWFSearch search plugin.

    This plugin, which inherits from :class:`~eodag.plugins.search.build_search_result.ECMWFSearch`,
    performs a POST request and uses its result to build a single :class:`~eodag.api.search_result.SearchResult`
    object. In contrast to ECMWFSearch or MeteoblueSearch, the products are only build with information
    returned by the provider.

    The available configuration parameters are inherited from parent classes, with some a particularity
    for pagination for this plugin.

    :param provider: An eodag providers configuration dictionary
    :param config: Search plugin configuration:

        * :attr:`~eodag.config.PluginConfig.pagination` (:class:`~eodag.config.PluginConfig.Pagination`)
          (**mandatory**): The configuration of how the pagination is done on the provider. For
          this plugin it has the node:

          * :attr:`~eodag.config.PluginConfig.Pagination.next_page_query_obj` (``str``): The
            additional parameters needed to perform search. These parameters won't be included in
            the result. This must be a json dict formatted like ``{{"foo":"bar"}}`` because it
            will be passed to a :meth:`str.format` method before being loaded as json.
    """

    def normalize_results(
        self, results: RawSearchResult, **kwargs: Any
    ) -> list[EOProduct]:
        """Build :class:`~eodag.api.product._product.EOProduct` from provider result

        :param results: Raw provider result as single dict in list
        :param kwargs: Search arguments
        :returns: list of single :class:`~eodag.api.product._product.EOProduct`
        """

        if kwargs.get("id") and "ORDERABLE" not in kwargs["id"]:
            # id is order id (only letters and numbers) -> use parent normalize results
            return super().normalize_results(results, **kwargs)

        # formating of eodag:order_link requires access to the collection value.
        results.data = [
            {**result, **results.collection_def_params} for result in results
        ]

        normalized = QueryStringSearch.normalize_results(self, results, **kwargs)

        if not normalized:
            return normalized

        # remove unwanted query params
        excluded_query_params = getattr(self.config, "remove_from_query", [])
        filtered_query_params = {
            k: v
            for k, v in results.query_params.items()
            if k not in excluded_query_params
        }
        for product in normalized:
            properties = {**product.properties, **results.query_params}
            properties["_dc_qs"] = quote_plus(orjson.dumps(filtered_query_params))
            product.properties = {ecmwf_format(k): v for k, v in properties.items()}

            # update product and title the same way as in parent class
            splitted_id = product.properties.get("title", "").split("-")
            dataset = "_".join(splitted_id[:-1])
            query_hash = splitted_id[-1]
            product.properties["title"] = product.properties["id"] = (
                (product.collection or dataset or self.provider).upper()
                + "_ORDERABLE_"
                + query_hash
            )

        return normalized

    def do_search(
        self, prep: PreparedSearch = PreparedSearch(limit=None), **kwargs: Any
    ) -> RawSearchResult:
        """Should perform the actual search request.

        :param args: arguments to be used in the search
        :param kwargs: keyword arguments to be used in the search
        :return: list containing the results from the provider in json format
        """
        if "id" in kwargs and "ORDERABLE" not in kwargs["id"]:
            # id is order id (only letters and numbers) -> use parent normalize results.
            # No real search. We fake it all, then check order status using given id
            raw_search_results = RawSearchResult([{}])
            raw_search_results.search_params = kwargs
            raw_search_results.query_params = (
                prep.query_params if hasattr(prep, "query_params") else {}
            )
            raw_search_results.collection_def_params = (
                prep.collection_def_params
                if hasattr(prep, "collection_def_params")
                else {}
            )
            return raw_search_results
        else:
            return QueryStringSearch.do_search(self, prep, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2025, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import re
from collections import UserDict, UserList
from typing import TYPE_CHECKING, Any, Optional, cast

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr
from pydantic import ValidationError as PydanticValidationError
from pydantic import model_validator
from pydantic_core import ErrorDetails, InitErrorDetails, PydanticCustomError
from stac_pydantic.collection import Extent, Provider, SpatialExtent, TimeInterval
from stac_pydantic.links import Links

from eodag.api.product.metadata_mapping import NOT_AVAILABLE
from eodag.types.queryables import CommonStacMetadata
from eodag.types.stac_metadata import create_stac_metadata_model
from eodag.utils import STAC_VERSION
from eodag.utils.env import is_env_var_true
from eodag.utils.exceptions import ValidationError
from eodag.utils.repr import dict_to_html_table

if TYPE_CHECKING:
    from pydantic import ModelWrapValidatorHandler
    from typing_extensions import Self

    from eodag.api.core import EODataAccessGateway
    from eodag.api.search_result import SearchResult
    from eodag.types.queryables import QueryablesDict

logger = logging.getLogger("eodag.api.collection")

RFC3339_PATTERN = (
    r"^(\d{4})-(\d{2})-(\d{2})"
    r"(?:T(\d{2}):(\d{2}):(\d{2})(\.\d+)?"
    r"(Z|([+-])(\d{2}):(\d{2}))?)?$"
)


class Collection(BaseModel):
    """A class representing a collection.

    A Collection object is used to describe a group of related :class:`~eodag.api.product._product.EOProduct` objects.
    """

    id: str = Field()
    title: Optional[str] = Field(default=None)
    description: str = Field(default=NOT_AVAILABLE)
    extent: Extent = Field(
        default=Extent(
            spatial=SpatialExtent(bbox=[[-180.0, -90.0, 180.0, 90.0]]),  # type: ignore
            temporal=TimeInterval(interval=[[None, None]]),
        ),
        description=(
            "The temporal extent of the collection, following the STAC specification for extent definition (e.g. "
            '{"spatial": {"bbox": [[-180.0, -90.0, 180.0, 90.0]]}, '
            '"temporal": {"interval": [["2024-06-10T12:00:00Z", None]]}}'
            "), with date/time strings in RFC 3339 format"
        ),
    )
    keywords: Optional[list[str]] = Field(default=None)
    license: str = Field(default="other")
    links: Optional[Links] = Field(default=None)
    providers: Optional[list[Provider]] = Field(default=None)

    # summaries
    constellation: Optional[str] = Field(default=None)
    instruments: Optional[list[str]] = Field(default=None)
    platform: Optional[str] = Field(default=None)
    processing_level: Optional[str] = Field(default=None, alias="processing:level")
    sci_doi: Optional[str] = Field(default=None, alias="sci:doi")
    eodag_sensor_type: Optional[str] = Field(default=None, alias="eodag:sensor_type")

    # eodag-specific attribute
    alias: Optional[str] = Field(
        default=None,
        description="An alias given by a user to use his customized id intead of the internal id of EODAG",
        repr=False,
    )

    # path to external collection metadata file (required by stac-fastapi-eodag)
    eodag_stac_collection: Optional[str] = Field(
        default=None, alias="stacCollection", exclude=True, repr=False
    )

    # Private property to store the eodag internal id value. Not part of the model schema.
    _id: str = PrivateAttr()
    _dag: Optional[EODataAccessGateway] = PrivateAttr(default=None)

    model_config = ConfigDict(
        extra="forbid", validate_by_name=True, serialize_by_alias=True
    )

    def model_post_init(self, context: Any) -> None:
        """Post-initialization method to set internal attributes."""
        self._id = self.id

    @classmethod
    def create_with_dag(cls, dag: EODataAccessGateway, **kwargs) -> Collection:
        """Create a Collection with a EODataAccessGateway instance.

        :param dag: The gateway instance to use to search products and to list queryables of the collection instance
        :param kwargs: The collection attributes
        """
        instance = cls(**kwargs)
        instance._dag = dag
        return instance

    @classmethod
    def get_collection_mtd_from_alias(cls, value: str) -> str:
        """Get collection metadata from alias

        >>> Collection.get_collection_mtd_from_alias('processing:level')
        'processing_level'
        """
        alias_map = {
            field_info.alias: name
            for name, field_info in cls.model_fields.items()
            if field_info.alias
        }
        return alias_map.get(value, value)

    @model_validator(mode="after")
    def set_id_from_alias(self) -> Self:
        """If an alias exists, use it to update ``id`` attribute"""
        if self.alias is not None:
            self._id = self.id
            self.id = self.alias
        return self

    @model_validator(mode="wrap")
    @classmethod
    def validate_collection(
        cls, values: dict[str, Any] | Self, handler: ModelWrapValidatorHandler[Self]
    ) -> Self:
        """Allow to create a collection instance with bad formatted attributes (except ``id``).
        Set incorrectly formatted attributes to ``None`` and ignore extra attributes.
        Log a warning about validation errors if ``EODAG_VALIDATE_COLLECTIONS`` environment variable is set to ``True``.
        """
        errors: list[ErrorDetails] = []
        continue_validation: bool = True

        # iterate over each step of validation where error(s) raise(s)
        while continue_validation:
            try:
                handler(values)
            except PydanticValidationError as e:
                tmp_errors = e.errors()
                # raise an error if the id is invalid
                if any(error["loc"][0] == "id" for error in tmp_errors):
                    raise ValidationError.from_error(e) from e

                # convert values to dict if it is a model instance
                values_dict = values if isinstance(values, dict) else values.__dict__

                # set incorrectly formatted attribute(s) to None and ignore its extra attribute(s)
                for error in tmp_errors:
                    wrong_param = error["loc"][0]
                    if not isinstance(wrong_param, str):
                        continue
                    if (
                        cls.get_collection_mtd_from_alias(wrong_param)
                        not in cls.model_fields
                    ):
                        del values_dict[wrong_param]
                    else:
                        values_dict[wrong_param] = cls.model_fields[
                            cls.get_collection_mtd_from_alias(wrong_param)
                        ].get_default()

                errors.extend(tmp_errors)
            else:
                continue_validation = False

        # log a warning if there were validation errors and the env var is set to True
        if errors and is_env_var_true("EODAG_VALIDATE_COLLECTIONS"):
            # log all errors at once
            error_title = f"collection {values_dict['id']}"
            init_errors: list[InitErrorDetails] = [
                InitErrorDetails(
                    type=PydanticCustomError(error["type"], error["msg"]),
                    loc=error["loc"],
                    input=error["input"],
                )
                for error in errors
            ]
            pydantic_error = PydanticValidationError.from_exception_data(
                title=error_title, line_errors=init_errors
            )
            logger.warning(pydantic_error)

        # Create a fresh instance with the cleaned values
        return handler(values)

    def __str__(self) -> str:
        return f'{type(self).__name__}("{self.id}")'

    def __repr_str__(self, join_str: str) -> str:
        return join_str.join(
            repr(v) if a is None else f"{a}={v!r}" for a, v in self.__repr_args__() if v
        )

    def _repr_html_(self, embedded: bool = False) -> str:
        thead = (
            f"""<thead><tr><td style='text-align: left; color: grey;'>
                {type(self).__name__}("<span style='color: black'>{self.id}</span>")</td></tr></thead>
            """
            if not embedded
            else ""
        )
        tr_style = "style='background-color: transparent;'" if embedded else ""
        col_html_table = dict_to_html_table(
            self.model_dump(exclude={"alias"}), depth=1, brackets=False
        )

        return (
            f"<table>{thead}<tbody>"
            f"<tr {tr_style}><td style='text-align: left;'>"
            f"{col_html_table}</td></tr>"
            "</tbody></table>"
        )

    def _ensure_dag(self) -> EODataAccessGateway:
        if self._dag is None:
            raise RuntimeError(
                f"Collection '{self.id}' needs EODataAccessGateway to perform this operation. "
                "Create with: Collection.create_with_dag(dag, id='...')"
            )
        return self._dag

    def search(self, **kwargs: Any) -> SearchResult:
        """Look for products of this collection matching criteria using the ``dag`` attribute of the instance.

        :param kwargs: Some other criteria that will be used to do the search,
                       using parameters compatible with the provider

        :returns: A collection of EO products matching the criteria.
        :raises: :class:`~eodag.utils.exceptions.ValidationError`: If the `collection` argument is set in `kwargs`,
                                                                   since it is already defined by the instance
        """
        dag = self._ensure_dag()
        collection_search_arg = "collection"
        if collection_search_arg in kwargs:
            raise ValidationError(
                f"{collection_search_arg} should not be set in kwargs since a collection instance is used",
                {collection_search_arg},
            )

        return dag.search(collection=self.id, **kwargs)

    def list_queryables(self, **kwargs: Any) -> QueryablesDict:
        """Fetch the queryable properties for this collection using the ``dag`` attribute of the instance.

        :param kwargs: additional filters for queryables

        :returns: A :class:`~eodag.api.product.queryables.QuerybalesDict` containing the EODAG queryable
                  properties, associating parameters to their annotated type, and an ``additional_properties`` attribute
        :raises: :class:`~eodag.utils.exceptions.ValidationError`: If the `collection` argument is set in `kwargs`,
                                                                   since it is already defined by the instance
        """
        dag = self._ensure_dag()
        collection_search_arg = "collection"
        if collection_search_arg in kwargs:
            raise ValidationError(
                f"{collection_search_arg} should not be set in kwargs since a collection instance is used",
                {collection_search_arg},
            )

        return dag.list_queryables(collection=self.id, **kwargs)

    def serialize(self) -> dict[str, Any]:
        """Serialize the Collection instance to a STAC dictionary.

        :returns: A STAC dictionary representation of the Collection instance.
        """
        stac_dict: dict[str, Any] = {
            "stac_version": STAC_VERSION,
            "type": "Collection",
        }

        stac_dict |= self.model_dump(mode="json", exclude_none=True, exclude={"alias"})

        stac_dict.setdefault("links", [])
        stac_dict.setdefault("providers", [])

        not_in_summaries = [
            "stac_version",
            "type",
            "id",
            "title",
            "description",
            "extent",
            "keywords",
            "license",
            "links",
            "providers",
        ]
        summaries = dict()
        for k, v in stac_dict.items():
            if k not in not_in_summaries:
                if isinstance(v, list):
                    summaries[k] = v
                elif isinstance(v, str):
                    summaries[k] = v.split(",")
                else:
                    summaries[k] = [v]
        stac_dict["summaries"] = summaries

        # Remove empty items and items moved to summaries
        keys_to_remove = [
            k
            for k in stac_dict.keys()
            if k not in not_in_summaries and k != "summaries"
        ]
        for k in keys_to_remove:
            del stac_dict[k]

        # add extensions
        summaries_model = cast(CommonStacMetadata, create_stac_metadata_model())
        summaries_validated = summaries_model.model_construct(
            _fields_set=None, **summaries
        )
        stac_dict["stac_extensions"] = summaries_validated.get_conformance_classes()

        return stac_dict


class CollectionsDict(UserDict[str, Collection]):
    """A UserDict object which values are :class:`~eodag.api.collection.Collection` objects, keyed by provider ``id``.

    :param collections: A list of collections

    :cvar data: List of collections
    """

    def __init__(
        self,
        collections: list[Collection],
    ) -> None:
        super().__init__()

        self.data = {col._id: col for col in collections}

    def __str__(self) -> str:
        return "{" + ", ".join(f'"{col}": {col_f}' for col, col_f in self.items()) + "}"

    def __repr__(self) -> str:
        return str(self)


class CollectionsList(UserList[Collection]):
    """An object representing a collection of :class:`~eodag.api.collection.Collection`.

    :param collections: A list of collections

    :cvar data: List of collections
    """

    def __init__(
        self,
        collections: list[Collection],
    ) -> None:
        super().__init__(collections)

    def __str__(self) -> str:
        return f"{type(self).__name__}([{', '.join(str(col) for col in self)}])"

    def __repr__(self) -> str:
        return str(self)

    def _repr_html_(self, embedded: bool = False) -> str:
        # mock "thead" tag by reproduicing its style to make "details" and "summary" tags work properly
        mock_thead = (
            f"""<details class='foldable'>
                <summary style='text-align: left; color: grey; font-size: 12px;'>
                {type(self).__name__}&ensp;({len(self)})
                </summary>
            """
            if not embedded
            else ""
        )
        tr_style = "style='background-color: transparent;'" if embedded else ""

        return (
            f"{mock_thead}<table><tbody>"
            + "".join(
                [
                    f"""<tr {tr_style}><td style='text-align: left;'>
                        <details>
                        <summary style='color: grey; font-family: monospace;'>
                        {i}&ensp;
                        {type(col).__name__}("<span style='color: black'>{col.id}</span>")
                    </summary>
                    {re.sub(r"(<thead>.*|.*</thead>)", "", col._repr_html_())}
                    </details>
                    </td></tr>
                    """
                    for i, col in enumerate(self)
                ]
            )
            + "</tbody></table></details>"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2025, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Compare two ext_product_types.json files and group differences by JSON paths.
"""

import json
import sys
from collections import defaultdict
from typing import Set


def get_all_paths(obj, path=""):
    """Get all JSON paths and their values from a nested object."""
    paths = {}

    if isinstance(obj, dict):
        for key, value in obj.items():
            new_path = f"{path}.{key}" if path else key
            if isinstance(value, (dict, list)):
                paths.update(get_all_paths(value, new_path))
            else:
                paths[new_path] = value
    elif isinstance(obj, list):
        for i, value in enumerate(obj):
            new_path = f"{path}[{i}]"
            if isinstance(value, (dict, list)):
                paths.update(get_all_paths(value, new_path))
            else:
                paths[new_path] = value
    else:
        paths[path] = obj

    return paths


def find_differing_paths(obj1, obj2) -> Set[str]:
    """Find all JSON paths where two objects differ."""
    paths1 = get_all_paths(obj1)
    paths2 = get_all_paths(obj2)

    all_paths = set(paths1.keys()) | set(paths2.keys())
    differing_paths = set()

    for path in all_paths:
        val1 = paths1.get(path, "<MISSING>")
        val2 = paths2.get(path, "<MISSING>")
        if val1 != val2:
            differing_paths.add(path)

    return differing_paths


def format_diff(obj1, obj2, indent=4):
    """Format a simple diff between two objects."""
    import difflib

    json1 = json.dumps(obj1, indent=indent, sort_keys=True)
    json2 = json.dumps(obj2, indent=indent, sort_keys=True)

    diff = list(
        difflib.unified_diff(
            json1.splitlines(keepends=True),
            json2.splitlines(keepends=True),
            fromfile="old",
            tofile="new",
            n=3,
        )
    )

    return "".join(diff)


def compare_product_types(
    file1_path: str, file2_path: str, include_details: bool = True, job_url: str = ""
):
    """Compare two ext_product_types.json files.

    Args:
        file1_path: Path to the old/reference JSON file
        file2_path: Path to the new JSON file
        include_details: Whether to include detailed diff content in collapsible sections
        job_url: URL to the job summary for linking from light version
    """  # Load JSON files
    try:
        with open(file1_path) as f:
            data1 = json.load(f)
        with open(file2_path) as f:
            data2 = json.load(f)
    except Exception as e:
        print(f"Error loading files: {e}", file=sys.stderr)
        return

    # 1. Compare providers (top-level keys)
    providers1 = set(data1.keys())
    providers2 = set(data2.keys())

    if providers1 != providers2:
        print("\n### Providers:")
        print("```diff")
        for provider in sorted(providers1 - providers2):
            print(f"- {provider}")
        for provider in sorted(providers2 - providers1):
            print(f"+ {provider}")
        print("```")

    # 2. Group differences by changed paths
    path_groups = defaultdict(list)  # path_set -> [(provider, type, item, diff)]

    # Compare each provider
    common_providers = providers1 & providers2
    for provider in sorted(common_providers):
        provider_data1 = data1[provider]
        provider_data2 = data2[provider]

        # Compare providers_config items content (no list diff, just content)
        providers_config1 = set(provider_data1.get("providers_config", {}).keys())
        providers_config2 = set(provider_data2.get("providers_config", {}).keys())

        # Compare each providers_config item's content
        common_provider_configs = providers_config1 & providers_config2
        for item in sorted(common_provider_configs):
            prov_conf1 = provider_data1["providers_config"][item]
            prov_conf2 = provider_data2["providers_config"][item]

            if prov_conf1 != prov_conf2:
                differing_paths = find_differing_paths(prov_conf1, prov_conf2)
                if differing_paths:
                    # Use sorted tuple of paths as key for grouping
                    path_key = tuple(sorted(differing_paths))
                    diff_content = format_diff(prov_conf1, prov_conf2)
                    path_groups[path_key].append(
                        (provider, "providers_config", item, diff_content)
                    )

        # Compare product_type lists
        product_types1 = set(provider_data1.get("product_types_config", {}).keys())
        product_types2 = set(provider_data2.get("product_types_config", {}).keys())

        if product_types1 != product_types2:
            print(f"\n#### {provider} - product_types:")
            print("```diff")
            for product_type in sorted(product_types1 - product_types2):
                print(f"- {product_type}")
            for product_type in sorted(product_types2 - product_types1):
                print(f"+ {product_type}")
            print("```")

        # Compare each product_type's content
        common_product_types = product_types1 & product_types2
        for product_type in sorted(common_product_types):
            coll1 = provider_data1["product_types_config"][product_type]
            coll2 = provider_data2["product_types_config"][product_type]

            if coll1 != coll2:
                differing_paths = find_differing_paths(coll1, coll2)
                if differing_paths:
                    # Use sorted tuple of paths as key for grouping
                    path_key = tuple(sorted(differing_paths))
                    diff_content = format_diff(coll1, coll2)
                    path_groups[path_key].append(
                        (provider, "product_types_config", product_type, diff_content)
                    )

    # 3. Display grouped results
    if path_groups:
        print("\n - - - \n")
        print("### Changes grouped by JSON paths:")

        for path_set, items in sorted(path_groups.items()):
            # Show the paths that changed
            paths_formatted = "`\n`".join(path_set)

            # Create anchor for this path group (for linking from light version)
            path_anchor = (
                "-".join(path_set)
                .replace(".", "")
                .replace("[", "")
                .replace("]", "")
                .replace("/", "-")
            )

            print(f'\n<a id="path-{path_anchor}"></a>')
            print(f"`{paths_formatted}`")

            # Get unique providers for this path group
            providers = sorted(set(item[0] for item in items))
            providers_text = f" ({', '.join(providers)})" if providers else ""

            # Show product_type count with providers
            print(f"**{len(items)} product_type(s) affected{providers_text}**")

            if include_details:
                # Make item list collapsible with detailed diffs
                print("<details>")
                print("<summary>Click to expand for detailed diffs</summary>")
                print()

                # Show each affected item
                for provider, config_type, item_name, diff_content in items:
                    print(f"##### {provider} - {config_type} - {item_name}")
                    print("```diff")
                    print(diff_content.rstrip())
                    print("```")
                    print()

                print("</details>")
            else:
                # Light version: show details if ≤10 product_types, otherwise link to detailed version
                if len(items) <= 10:
                    # Show collapsible details like in full version
                    print("<details>")
                    print("<summary>Click to expand for detailed diffs</summary>")
                    print()

                    # Show each affected item with diffs
                    for provider, config_type, item_name, diff_content in items:
                        print(f"##### {provider} - {config_type} - {item_name}")
                        print("```diff")
                        print(diff_content.rstrip())
                        print("```")
                        print()

                    print("</details>")
                else:
                    # Too many items, just link to the detailed anchor in job summary
                    print(
                        "\n_Too many product_types to list here. See "
                        f"[detailed breakdown in job summary]({job_url}#path-{path_anchor}) "
                        "for the complete list._"
                    )
                    print()

            print("\n - - - \n")

    # Summary
    total_changes = sum(len(items) for items in path_groups.values())
    total_paths = len(path_groups)

    if total_changes == 0:
        print("### Summary:")
        print("- No changes detected in product-types configuration")
    else:
        print("### Summary:")
        print(
            f"- {total_changes} item(s) with changes (providers_config + product_types_config)"
        )
        print(f"- {total_paths} unique path pattern(s)")


def main():
    """Main entry point for the product_types comparison script."""
    if len(sys.argv) < 3:
        print(
            "Usage: python ext_product_types_cmp.py <old.json> <new.json> [--no-details] [--job-url=URL]"
        )
        sys.exit(1)

    include_details = "--no-details" not in sys.argv

    # Extract job URL if provided
    job_url = ""
    for arg in sys.argv:
        if arg.startswith("--job-url="):
            job_url = arg.split("=", 1)[1]
            break

    compare_product_types(sys.argv[1], sys.argv[2], include_details, job_url)


if __name__ == "__main__":
    main()
